import pytest

from integration.conftest import ClientFactory
from weaviate.auth import Auth
from weaviate.rbac.models import GroupTypes, Role

RBAC_PORTS = (8092, 50063)
RBAC_AUTH_CREDS = Auth.api_key("admin-key")


def test_assign_and_get_group_roles_oidc(client_factory: ClientFactory) -> None:
    with client_factory(ports=RBAC_PORTS, auth_credentials=RBAC_AUTH_CREDS) as client:
        if client._connection._weaviate_version.is_lower_than(1, 32, 0):
            pytest.skip("This test requires Weaviate 1.32.0 or higher")

        roles_to_assign = ["viewer", "admin"]
        group_id = "/assign-group"
        client.groups.oidc.revoke_roles(group_id=group_id, role_names=roles_to_assign)

        roles_base = client.groups.oidc.get_assigned_roles(
            group_id=group_id,
        )
        assert len(roles_base) == 0

        client.groups.oidc.assign_roles(group_id=group_id, role_names=roles_to_assign)

        roles = client.groups.oidc.get_assigned_roles(group_id=group_id, include_permissions=True)
        assert len(roles) == 2
        for role_name in roles_to_assign:
            assert role_name in roles
            assert isinstance(roles[role_name], Role)

        client.groups.oidc.revoke_roles(group_id=group_id, role_names=roles_to_assign)

        roles_base = client.groups.oidc.get_assigned_roles(
            group_id=group_id,
        )
        assert len(roles_base) == 0


def test_known_groups(client_factory: ClientFactory) -> None:
    with client_factory(ports=RBAC_PORTS, auth_credentials=RBAC_AUTH_CREDS) as client:
        if client._connection._weaviate_version.is_lower_than(1, 32, 0):
            pytest.skip("This test requires Weaviate 1.32.0 or higher")

        group1 = "/known-group1"
        group2 = "/known-group2"
        client.groups.oidc.assign_roles(group_id=group1, role_names="viewer")
        client.groups.oidc.assign_roles(group_id=group2, role_names="viewer")

        groups = client.groups.oidc.get_known_group_names()
        assert len(groups) >= 2  # other tests may add groups
        assert group1 in groups
        assert group2 in groups

        client.groups.oidc.revoke_roles(group_id=group1, role_names="viewer")
        client.groups.oidc.revoke_roles(group_id=group2, role_names="viewer")

        groups = client.groups.oidc.get_known_group_names()
        assert len(groups) >= 0  # other tests may add groups
        assert group1 not in groups
        assert group2 not in groups


def test_get_group_assignments(client_factory: ClientFactory) -> None:
    with client_factory(ports=RBAC_PORTS, auth_credentials=RBAC_AUTH_CREDS) as client:
        if client._connection._weaviate_version.is_lower_than(1, 32, 0):
            pytest.skip("This test requires Weaviate 1.32.0 or higher")

        role_name = "test_group_assignments_role"
        client.roles.delete(role_name=role_name)
        client.roles.create(role_name=role_name, permissions=[])

        group_assignments = client.roles.get_group_assignments(role_name=role_name)
        assert len(group_assignments) == 0

        client.groups.oidc.assign_roles(group_id="custom-group", role_names=role_name)
        client.groups.oidc.assign_roles(group_id="custom-group2", role_names=role_name)

        group_assignments = client.roles.get_group_assignments(role_name=role_name)
        assert len(group_assignments) == 2
        for group in group_assignments:
            assert group.group_id in ["custom-group", "custom-group2"]
            assert group.group_type == GroupTypes.OIDC

        client.groups.oidc.revoke_roles(group_id="custom-group", role_names=role_name)
        client.groups.oidc.revoke_roles(group_id="custom-group2", role_names=role_name)



import uuid
from dataclasses import dataclass
from typing import Any, Generator, Optional, Protocol, Union

import pytest

from integration.conftest import CollectionFactory, CollectionFactoryGet
from weaviate.collections import Collection
from weaviate.collections.classes.config import (
    Configure,
    DataType,
    Property,
    ReferenceProperty,
)
from weaviate.collections.classes.grpc import QueryReference
from weaviate.collections.classes.internal import ReferenceToMulti, _CrossReference
from weaviate.collections.classes.tenants import Tenant
from weaviate.types import VECTORS

UUID = Union[str, uuid.UUID]


@dataclass
class MockNumpyTorch:
    array: list

    def squeeze(self) -> "MockNumpyTorch":
        return self

    def tolist(self) -> list:
        return self.array


@dataclass
class MockTensorFlow:
    array: list

    def numpy(self) -> "MockNumpyTorch":
        return MockNumpyTorch(self.array)


UUID1 = uuid.UUID("806827e0-2b31-43ca-9269-24fa95a221f9")
UUID2 = uuid.UUID("8ad0d33c-8db1-4437-87f3-72161ca2a51a")
UUID3 = uuid.UUID("83d99755-9deb-4b16-8431-d1dff4ab0a75")
UUID4 = uuid.UUID("385c992b-452a-4f71-a5d9-b161f51b540d")
UUID5 = uuid.UUID("0a4d16b3-c418-40d3-a6b7-51f87c7a603b")
UUID6 = uuid.UUID("c8a201b6-fdd2-48d1-a8ee-289a540b1b4b")


class BatchCollection(Protocol):
    """Typing for fixture."""

    def __call__(self, name: str = "", multi_tenancy: bool = False) -> Collection[Any, Any]:
        """Typing for fixture."""
        ...


@pytest.fixture
def batch_collection(
    collection_factory: CollectionFactory,
) -> Generator[BatchCollection, None, None]:
    def _factory(name: str = "", multi_tenancy: bool = False) -> Collection[Any, Any]:
        collection = collection_factory(
            name=name,
            vectorizer_config=Configure.Vectorizer.none(),
            properties=[
                Property(name="name", data_type=DataType.TEXT),
                Property(name="age", data_type=DataType.INT),
            ],
            multi_tenancy_config=Configure.multi_tenancy(multi_tenancy),
        )
        collection.config.add_reference(
            ReferenceProperty(name="test", target_collection=collection.name)
        )

        return collection

    yield _factory


@pytest.mark.parametrize(
    "vector",
    [None, [1, 2, 3], MockNumpyTorch([1, 2, 3]), MockTensorFlow([1, 2, 3])],
)
@pytest.mark.parametrize("uuid", [None, UUID1, str(UUID2), UUID3.hex])
def test_add_object(
    batch_collection: BatchCollection, uuid: Optional[UUID], vector: Optional[VECTORS]
) -> None:
    collection = batch_collection()

    with collection.batch.dynamic() as batch:
        batch.add_object(uuid=uuid, vector=vector)
    assert len(collection.batch.failed_objects) == 0
    assert len(collection.batch.failed_references) == 0
    objs = collection.query.fetch_objects().objects
    assert len(objs) == 1


@pytest.mark.parametrize("from_uuid", [UUID1, str(UUID2), UUID3.hex])
@pytest.mark.parametrize("to_uuid", [UUID4.hex, UUID5, str(UUID6)])
def test_add_reference(
    batch_collection: BatchCollection,
    from_uuid: UUID,
    to_uuid: UUID,
) -> None:
    """Test the `add_reference` method."""
    collection = batch_collection()

    with collection.batch.dynamic() as batch:
        batch.add_object(uuid=from_uuid)
        batch.add_object(uuid=to_uuid)
        batch.add_reference(from_uuid=from_uuid, from_property="test", to=to_uuid)
    assert len(collection.batch.failed_objects) == 0
    assert len(collection.batch.failed_references) == 0
    objs = collection.query.fetch_objects().objects
    obj = collection.query.fetch_object_by_id(
        from_uuid, return_references=QueryReference(link_on="test")
    )
    assert len(objs) == 2
    assert isinstance(obj.references["test"], _CrossReference)


def test_add_object_batch_with_tenant(batch_collection: BatchCollection) -> None:
    mt_collection = batch_collection(multi_tenancy=True)

    mt_collection.tenants.create([Tenant(name="tenant" + str(i)) for i in range(5)])
    for i in range(5):
        col = mt_collection.with_tenant("tenant" + str(i % 5))
        with col.batch.fixed_size(batch_size=10, concurrent_requests=1) as batch:
            batch.add_object(properties={"name": "tenant" + str(i % 5)})
        assert len(col.batch.failed_objects) == 0
        assert len(col.batch.failed_references) == 0
    objs = mt_collection.with_tenant("tenant1").query.fetch_objects().objects
    assert len(objs) == 1
    for obj in objs:
        assert obj.properties["name"] == "tenant1"


def test_add_ref_batch_with_tenant(batch_collection: BatchCollection) -> None:
    mt_collection = batch_collection(multi_tenancy=True)

    mt_collection.tenants.create([Tenant(name="tenant" + str(i)) for i in range(5)])

    batching = mt_collection.with_tenant("tenant1")
    with batching.batch.dynamic() as batch:
        obj_uuid0 = uuid.uuid4()
        batch.add_object(properties={"name": "one"}, uuid=obj_uuid0)

        obj_uuid1 = uuid.uuid4()
        batch.add_object(
            properties={"name": "two"},
            uuid=obj_uuid1,
        )

        # add refs between classes for all tenants
        batch.add_reference(
            from_property="test",
            from_uuid=obj_uuid1,
            to=ReferenceToMulti(uuids=obj_uuid0, target_collection=mt_collection.name),
        )
        batch.add_reference(
            from_property="test",
            from_uuid=obj_uuid0,
            to=ReferenceToMulti(uuids=obj_uuid1, target_collection=mt_collection.name),
        )
        # target collection required when inserting references into multi-tenancy collections
    assert len(batching.batch.failed_objects) == 0
    assert len(batching.batch.failed_references) == 0
    ret_obj = mt_collection.with_tenant("tenant1").query.fetch_object_by_id(
        obj_uuid0, return_references=QueryReference(link_on="test")
    )
    assert ret_obj.properties["name"] == "one"
    assert isinstance(ret_obj.references["test"], _CrossReference)
    assert ret_obj.references["test"].objects[0].uuid == obj_uuid1


def test_error_reset(batch_collection: BatchCollection) -> None:
    col = batch_collection()
    with col.batch.fixed_size(1) as batch:
        batch.add_object(properties={"name": 1})
        batch.add_object(properties={"name": "correct"})

        # make sure that errors are processed
        batch.flush()
        assert len(col.batch.failed_objects) == 0  # error is still private
        assert batch.number_errors == 1

    errs = col.batch.failed_objects
    assert len(errs) == 1
    assert errs[0].object_.properties is not None
    assert errs[0].object_.properties["name"] == 1

    with col.batch.dynamic() as batch:
        batch.add_object(properties={"name": 2})
        batch.add_object(properties={"name": "correct"})

    errs2 = col.batch.failed_objects
    assert len(errs2) == 1
    assert errs2[0].object_.properties is not None
    assert errs2[0].object_.properties["name"] == 2

    # err still contains original errors
    assert len(errs) == 1
    assert errs[0].object_.properties is not None
    assert errs[0].object_.properties["name"] == 1


def test_refs_and_objects(batch_collection: BatchCollection) -> None:
    """Test that references are not added before the source object is added."""
    col = batch_collection()
    uuids = [uuid.uuid4() for _ in range(10)]
    with col.batch.fixed_size(1, concurrent_requests=1) as batch:
        for uid in uuids:
            batch.add_object(properties={}, uuid=uid)
        batch.add_reference(
            from_uuid=uuids[-1],
            from_property="test",
            to=uuids[-1],
        )

    assert len(col.batch.failed_objects) == 0
    assert len(col.batch.failed_references) == 0

    obj = col.query.fetch_object_by_id(uuids[-1], return_references=QueryReference(link_on="test"))
    assert "test" in obj.references
    assert obj.references["test"].objects[0].uuid == uuids[-1]


def test_non_existant_collection(collection_factory_get: CollectionFactoryGet) -> None:
    collection = collection_factory_get("DoesNotExist")
    with collection.batch.dynamic() as batch:
        batch.add_object(properties={"name": 2})

    # above should not throw - depending on the autoschema config this might create an error or
    # not, so we do not check for errors here



from typing import Generic, Iterable, Literal, Optional, Type, Union, overload

from weaviate.collections.classes.grpc import METADATA, PROPERTIES, REFERENCES, Sorting
from weaviate.collections.classes.internal import (
    CrossReferences,
    QueryReturn,
    QueryReturnType,
    ReturnProperties,
    ReturnReferences,
)
from weaviate.collections.classes.types import Properties, References, TProperties, TReferences
from weaviate.connect.v4 import ConnectionAsync
from weaviate.types import INCLUDE_VECTOR, UUID

from .executor import _FetchObjectsByIDsQueryExecutor

class _FetchObjectsByIDsQueryAsync(
    Generic[Properties, References],
    _FetchObjectsByIDsQueryExecutor[ConnectionAsync, Properties, References],
):
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Literal[None] = None,
    ) -> QueryReturn[Properties, References]: ...
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: REFERENCES,
    ) -> QueryReturn[Properties, CrossReferences]: ...
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Type[TReferences],
    ) -> QueryReturn[Properties, TReferences]: ...
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Literal[None] = None,
    ) -> QueryReturn[TProperties, References]: ...
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: REFERENCES,
    ) -> QueryReturn[TProperties, CrossReferences]: ...
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Type[TReferences],
    ) -> QueryReturn[TProperties, TReferences]: ...
    @overload
    async def fetch_objects_by_ids(
        self,
        ids: Iterable[UUID],
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> QueryReturnType[Properties, References, TProperties, TReferences]: ...



from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync
from weaviate.groups.base import _GroupsOIDCExecutor


@executor.wrap("async")
class _GroupsAsync:
    def __init__(self, connection: ConnectionAsync):
        self.oidc = _GroupsOIDCAsync(connection)


@executor.wrap("sync")
class _GroupsOIDCAsync(_GroupsOIDCExecutor[ConnectionAsync]):
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/tenants.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10v1/tenants.proto\x12\x0bweaviate.v1\"\\\n\x11TenantsGetRequest\x12\x12\n\ncollection\x18\x01 \x01(\t\x12)\n\x05names\x18\x02 \x01(\x0b\x32\x18.weaviate.v1.TenantNamesH\x00\x42\x08\n\x06params\"\x1d\n\x0bTenantNames\x12\x0e\n\x06values\x18\x01 \x03(\t\"E\n\x0fTenantsGetReply\x12\x0c\n\x04took\x18\x01 \x01(\x02\x12$\n\x07tenants\x18\x02 \x03(\x0b\x32\x13.weaviate.v1.Tenant\"R\n\x06Tenant\x12\x0c\n\x04name\x18\x01 \x01(\t\x12:\n\x0f\x61\x63tivity_status\x18\x02 \x01(\x0e\x32!.weaviate.v1.TenantActivityStatus*\xaf\x03\n\x14TenantActivityStatus\x12&\n\"TENANT_ACTIVITY_STATUS_UNSPECIFIED\x10\x00\x12\x1e\n\x1aTENANT_ACTIVITY_STATUS_HOT\x10\x01\x12\x1f\n\x1bTENANT_ACTIVITY_STATUS_COLD\x10\x02\x12!\n\x1dTENANT_ACTIVITY_STATUS_FROZEN\x10\x04\x12%\n!TENANT_ACTIVITY_STATUS_UNFREEZING\x10\x05\x12#\n\x1fTENANT_ACTIVITY_STATUS_FREEZING\x10\x06\x12!\n\x1dTENANT_ACTIVITY_STATUS_ACTIVE\x10\x07\x12#\n\x1fTENANT_ACTIVITY_STATUS_INACTIVE\x10\x08\x12$\n TENANT_ACTIVITY_STATUS_OFFLOADED\x10\t\x12%\n!TENANT_ACTIVITY_STATUS_OFFLOADING\x10\n\x12$\n TENANT_ACTIVITY_STATUS_ONLOADING\x10\x0b\"\x04\x08\x03\x10\x03\x42q\n#io.weaviate.client.grpc.protocol.v1B\x14WeaviateProtoTenantsZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.tenants_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n#io.weaviate.client.grpc.protocol.v1B\024WeaviateProtoTenantsZ4github.com/weaviate/weaviate/grpc/generated;protocol'
  _globals['_TENANTACTIVITYSTATUS']._serialized_start=314
  _globals['_TENANTACTIVITYSTATUS']._serialized_end=745
  _globals['_TENANTSGETREQUEST']._serialized_start=33
  _globals['_TENANTSGETREQUEST']._serialized_end=125
  _globals['_TENANTNAMES']._serialized_start=127
  _globals['_TENANTNAMES']._serialized_end=156
  _globals['_TENANTSGETREPLY']._serialized_start=158
  _globals['_TENANTSGETREPLY']._serialized_end=227
  _globals['_TENANT']._serialized_start=229
  _globals['_TENANT']._serialized_end=311
# @@protoc_insertion_point(module_scope)



from typing import List

from flask import Flask, g

from journey_tests.journeys import SyncJourneys

app = Flask(__name__)


def get_sync_client() -> SyncJourneys:
    if "sync" not in g:
        g.sync = SyncJourneys.use()
    return g.sync


@app.route("/sync-in-sync")
def sync() -> List[dict]:
    return get_sync_client().simple()


if __name__ == "__main__":
    app.run(debug=False, host="0.0.0.0")



from __future__ import annotations

import time
from copy import copy
from dataclasses import dataclass, field
from ssl import SSLZeroReturnError
from threading import Event, Thread
from typing import (
    Any,
    Awaitable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from authlib.integrations.httpx_client import (  # type: ignore
    AsyncOAuth2Client,
    OAuth2Client,
)
from grpc import Call, RpcError, StatusCode
from grpc import Channel as SyncChannel  # type: ignore
from grpc.aio import AioRpcError
from grpc.aio import Channel as AsyncChannel  # type: ignore

# from grpclib.client import Channel
from httpx import (
    AsyncClient,
    AsyncHTTPTransport,
    Client,
    ConnectError,
    HTTPError,
    HTTPStatusError,
    HTTPTransport,
    Limits,
    Proxy,
    ReadError,
    ReadTimeout,
    RemoteProtocolError,
    RequestError,
    Response,
    Timeout,
)

from weaviate import __version__ as client_version
from weaviate.auth import AuthApiKey, AuthClientCredentials, AuthCredentials
from weaviate.config import ConnectionConfig, Proxies
from weaviate.config import Timeout as TimeoutConfig
from weaviate.connect import executor
from weaviate.connect.authentication import _Auth
from weaviate.connect.base import (
    ConnectionParams,
    JSONPayload,
    _get_proxies,
)
from weaviate.connect.event_loop import _EventLoopSingleton
from weaviate.connect.integrations import _IntegrationConfig
from weaviate.embedded import EmbeddedV4
from weaviate.exceptions import (
    AuthenticationFailedError,
    InsufficientPermissionsError,
    UnexpectedStatusCodeError,
    WeaviateBatchError,
    WeaviateClosedClientError,
    WeaviateConnectionError,
    WeaviateDeleteManyError,
    WeaviateGRPCUnavailableError,
    WeaviateInvalidInputError,
    WeaviateQueryError,
    WeaviateRetryError,
    WeaviateStartUpError,
    WeaviateTenantGetError,
    WeaviateTimeoutError,
)
from weaviate.proto.v1 import (
    aggregate_pb2,
    batch_delete_pb2,
    batch_pb2,
    health_weaviate_pb2,
    search_get_pb2,
    tenants_pb2,
    weaviate_pb2_grpc,
)
from weaviate.retry import _Retry
from weaviate.util import (
    PYPI_PACKAGE_URL,
    _decode_json_response_dict,
    _ServerVersion,
    is_weaviate_client_too_old,
    is_weaviate_domain,
)
from weaviate.validator import _validate_input, _ValidateArgument
from weaviate.warnings import _Warnings

Session = Union[Client, OAuth2Client]
AsyncSession = Union[AsyncClient, AsyncOAuth2Client]
HttpClient = Union[AsyncClient, AsyncOAuth2Client, Client, OAuth2Client]

PERMISSION_DENIED = "PERMISSION_DENIED"


@dataclass
class _ExpectedStatusCodes:
    ok_in: Union[List[int], int]
    error: str
    ok: List[int] = field(init=False)

    def __post_init__(self) -> None:
        if isinstance(self.ok_in, int):
            self.ok = [self.ok_in]
        else:
            self.ok = self.ok_in


class _ConnectionBase:
    def __init__(
        self,
        connection_params: ConnectionParams,
        auth_client_secret: Optional[AuthCredentials],
        timeout_config: TimeoutConfig,
        proxies: Union[str, Proxies, None],
        trust_env: bool,
        additional_headers: Optional[Dict[str, Any]],
        connection_config: ConnectionConfig,
        embedded_db: Optional[EmbeddedV4] = None,
        skip_init_checks: bool = False,
    ):
        self.url = connection_params._http_url
        self.embedded_db = embedded_db
        self._api_version_path = "/v1"
        self.__additional_headers = {}
        self._auth = auth_client_secret
        self._client: Optional[HttpClient] = None
        self._connection_params = connection_params
        self._grpc_stub: Optional[weaviate_pb2_grpc.WeaviateStub] = None
        self._grpc_channel: Union[AsyncChannel, SyncChannel, None] = None
        self.timeout_config = timeout_config
        self.__connection_config = connection_config
        self.__trust_env = trust_env
        self._weaviate_version = _ServerVersion.from_string("")
        self._grpc_max_msg_size: Optional[int] = None
        self._connected = False
        self._skip_init_checks = skip_init_checks

        self._headers = {"content-type": "application/json"}
        self.__add_weaviate_embedding_service_header(connection_params.http.host)
        if additional_headers is not None:
            _validate_input(_ValidateArgument([dict], "additional_headers", additional_headers))
            self.__additional_headers = additional_headers
            for key, value in additional_headers.items():
                if value is None:
                    raise WeaviateInvalidInputError(
                        f"Value for key '{key}' in headers cannot be None."
                    )
                self._headers[key.lower()] = value

        self._proxies: Dict[str, str] = _get_proxies(proxies, trust_env)

        # auth secrets can contain more information than a header (refresh tokens and lifetime) and therefore take
        # precedent over headers
        if "authorization" in self._headers and auth_client_secret is not None:
            _Warnings.auth_header_and_auth_secret()
            self._headers.pop("authorization")

        # if there are API keys included add them right away to headers
        if auth_client_secret is not None and isinstance(auth_client_secret, AuthApiKey):
            self._headers["authorization"] = "Bearer " + auth_client_secret.api_key

        self._prepare_grpc_headers()

    def __add_weaviate_embedding_service_header(self, wcd_host: str) -> None:
        if not is_weaviate_domain(wcd_host):
            return

        self._headers["X-Weaviate-Cluster-URL"] = "https://" + wcd_host
        if isinstance(self._auth, AuthApiKey):
            # keeping for backwards compatibility for older clusters for now. On newer clusters, Embedding Service reuses Authorization header.
            self._headers["X-Weaviate-Api-Key"] = self._auth.api_key

    def set_integrations(self, integrations_config: List[_IntegrationConfig]) -> None:
        for integration in integrations_config:
            self._headers.update(integration._to_header())
            self.__additional_headers.update(integration._to_header())

    @overload
    def _make_client(self, colour: Literal["async"]) -> AsyncClient: ...

    @overload
    def _make_client(self, colour: Literal["sync"]) -> Client: ...

    def _make_client(self, colour: executor.Colour) -> Union[AsyncClient, Client]:
        if colour == "async":
            return AsyncClient(
                headers=self._headers,
                mounts=self._make_mounts(colour),
                trust_env=self.__trust_env,
            )
        if colour == "sync":
            return Client(
                headers=self._headers,
                mounts=self._make_mounts(colour),
                trust_env=self.__trust_env,
            )

    @overload
    def _make_mounts(self, colour: Literal["async"]) -> Dict[str, AsyncHTTPTransport]: ...

    @overload
    def _make_mounts(self, colour: Literal["sync"]) -> Dict[str, HTTPTransport]: ...

    def _make_mounts(
        self, colour: executor.Colour
    ) -> Union[Dict[str, AsyncHTTPTransport], Dict[str, HTTPTransport]]:
        if colour == "async":
            return {
                (f"{key}://" if key == "http" or key == "https" else key): AsyncHTTPTransport(
                    limits=Limits(
                        max_connections=self.__connection_config.session_pool_maxsize,
                        max_keepalive_connections=self.__connection_config.session_pool_connections,
                    ),
                    proxy=Proxy(url=proxy),
                    retries=self.__connection_config.session_pool_max_retries,
                    trust_env=self.__trust_env,
                )
                for key, proxy in self._proxies.items()
                if key != "grpc"
            }
        if colour == "sync":
            return {
                f"{key}://" if key == "http" or key == "https" else key: HTTPTransport(
                    limits=Limits(
                        max_connections=self.__connection_config.session_pool_maxsize,
                        max_keepalive_connections=self.__connection_config.session_pool_connections,
                    ),
                    proxy=Proxy(url=proxy),
                    retries=self.__connection_config.session_pool_max_retries,
                    trust_env=self.__trust_env,
                )
                for key, proxy in self._proxies.items()
                if key != "grpc"
            }

    def is_connected(self) -> bool:
        return self._connected

    def get_current_bearer_token(self) -> str:
        if "authorization" in self._headers:
            return self._headers["authorization"]
        elif isinstance(self._client, (OAuth2Client, AsyncOAuth2Client)):
            return f"Bearer {self._client.token['access_token']}"
        return ""

    def _prepare_grpc_headers(self) -> None:
        self.__metadata_list: List[Tuple[str, str]] = []
        if len(self.additional_headers):
            for key, val in self.additional_headers.items():
                if val is not None:
                    self.__metadata_list.append((key.lower(), val))

        if self._auth is not None:
            if "X-Weaviate-Cluster-URL" in self._headers:
                self.__metadata_list.append(
                    ("x-weaviate-cluster-url", self._headers["X-Weaviate-Cluster-URL"])
                )

            if isinstance(self._auth, AuthApiKey):
                if "X-Weaviate-Api-Key" in self._headers:
                    # keeping for backwards compatibility for older clusters for now. On newer clusters, Embedding Service reuses Authorization header.
                    self.__metadata_list.append(
                        ("x-weaviate-api-key", self._headers["X-Weaviate-Api-Key"])
                    )
                self.__metadata_list.append(("authorization", "Bearer " + self._auth.api_key))
            else:
                self.__add_weaviate_embedding_service_auth_grpc_header()
                self.__metadata_list.append(
                    ("authorization", "dummy_will_be_refreshed_for_each_call")
                )

        if len(self.__metadata_list) > 0:
            self.__grpc_headers: Optional[Tuple[Tuple[str, str], ...]] = tuple(self.__metadata_list)
        else:
            self.__grpc_headers = None

    def __add_weaviate_embedding_service_auth_grpc_header(self) -> None:
        if is_weaviate_domain(self._connection_params.http.host):
            # keeping for backwards compatibility for older clusters for now. On newer clusters, Embedding Service reuses Authorization header.
            self.__metadata_list.append(
                ("x-weaviate-api-key", "dummy_will_be_refreshed_for_each_call")
            )

    def grpc_headers(self) -> Optional[Tuple[Tuple[str, str], ...]]:
        if self._auth is None or isinstance(self._auth, AuthApiKey):
            return self.__grpc_headers

        assert self.__grpc_headers is not None
        access_token = self.get_current_bearer_token()
        self.__refresh_weaviate_embedding_service_auth_grpc_header()
        # auth is last entry in list, rest is static
        self.__metadata_list[len(self.__metadata_list) - 1] = (
            "authorization",
            access_token,
        )
        return tuple(self.__metadata_list)

    def __refresh_weaviate_embedding_service_auth_grpc_header(self) -> None:
        if is_weaviate_domain(self._connection_params.http.host):
            # keeping for backwards compatibility for older clusters for now. On newer clusters, Embedding Service reuses Authorization header.
            self.__metadata_list[len(self.__metadata_list) - 2] = (
                "x-weaviate-api-key",
                self.get_current_bearer_token(),
            )

    def _ping_grpc(self, colour: executor.Colour) -> Union[None, Awaitable[None]]:
        """Performs a grpc health check and raises WeaviateGRPCUnavailableError if not."""
        assert self._grpc_channel is not None

        try:
            res = self._grpc_channel.unary_unary(
                "/grpc.health.v1.Health/Check",
                request_serializer=health_weaviate_pb2.WeaviateHealthCheckRequest.SerializeToString,
                response_deserializer=health_weaviate_pb2.WeaviateHealthCheckResponse.FromString,
            )(health_weaviate_pb2.WeaviateHealthCheckRequest(), timeout=self.timeout_config.init)
            if colour == "async":

                async def execute():
                    assert isinstance(res, Awaitable)
                    try:
                        return self.__handle_ping_response(
                            cast(health_weaviate_pb2.WeaviateHealthCheckResponse, await res)
                        )
                    except Exception as e:
                        self.__handle_ping_exception(e)

                return execute()
            assert not isinstance(res, Awaitable)
            self.__handle_ping_response(cast(health_weaviate_pb2.WeaviateHealthCheckResponse, res))
            return None

        except Exception as e:
            self.__handle_ping_exception(e)
        return None

    def __handle_ping_response(self, res: health_weaviate_pb2.WeaviateHealthCheckResponse) -> None:
        if res.status != health_weaviate_pb2.WeaviateHealthCheckResponse.SERVING:
            raise WeaviateGRPCUnavailableError(
                f"v{self.server_version}", self._connection_params._grpc_address
            )
        return None

    def __handle_ping_exception(self, e: Exception) -> None:
        raise WeaviateGRPCUnavailableError(
            f"v{self.server_version}", self._connection_params._grpc_address
        ) from e

    @property
    def grpc_stub(self) -> Optional[weaviate_pb2_grpc.WeaviateStub]:
        if not self.is_connected():
            raise WeaviateClosedClientError()
        return self._grpc_stub

    def __del__(self) -> None:
        if self._client is not None or self._grpc_channel is not None:
            _Warnings.unclosed_connection()

    @property
    def server_version(self) -> str:
        """Version of the weaviate instance."""
        return str(self._weaviate_version)

    def get_proxies(self) -> Dict[str, str]:
        return self._proxies

    @property
    def additional_headers(self) -> Dict[str, str]:
        return self.__additional_headers

    def __make_clients(self, colour: Literal["async", "sync"]) -> None:
        self._client = self._make_client(colour)

    def open_connection_grpc(self, colour: executor.Colour) -> None:
        channel = self._connection_params._grpc_channel(
            proxies=self._proxies,
            grpc_msg_size=self._grpc_max_msg_size,
            is_async=colour == "async",
        )
        self._grpc_channel = channel
        assert self._grpc_channel is not None
        self._grpc_stub = weaviate_pb2_grpc.WeaviateStub(self._grpc_channel)

    def _open_connections_rest(
        self, auth_client_secret: Optional[AuthCredentials], colour: executor.Colour
    ) -> Union[None, Awaitable[None]]:
        # API keys are separate from OIDC and do not need any config from weaviate
        if auth_client_secret is not None and isinstance(auth_client_secret, AuthApiKey):
            self.__make_clients(colour)
            return executor.empty(colour)

        if "authorization" in self._headers and auth_client_secret is None:
            self.__make_clients(colour)
            return executor.empty(colour)

        # no need to check OIDC if no auth is provided and users dont want any checks at initialization time
        if self._skip_init_checks and auth_client_secret is None:
            self.__make_clients(colour)
            return executor.empty(colour)

        oidc_url = self.url + self._api_version_path + "/.well-known/openid-configuration"
        if colour == "async":

            async def get_oidc() -> None:
                async with self._make_client("async") as client:
                    try:
                        response = await client.get(oidc_url)
                    except Exception as e:
                        raise WeaviateConnectionError(
                            f"Error: {e}. \nIs Weaviate running and reachable at {self.url}?"
                        )
                res = self.__process_oidc_response(response, auth_client_secret, oidc_url, colour)
                if isinstance(res, Awaitable):
                    return await res
                else:
                    return res

            return get_oidc()

        with self._make_client("sync") as client:
            try:
                response = client.get(oidc_url)
            except Exception as e:
                raise WeaviateConnectionError(
                    f"Error: {e}. \nIs Weaviate running and reachable at {self.url}?"
                )
        res = self.__process_oidc_response(response, auth_client_secret, oidc_url, colour)
        assert not isinstance(res, Awaitable)
        return res

    def __process_oidc_response(
        self,
        response: Response,
        auth_client_secret: Optional[AuthCredentials],
        oidc_url: str,
        colour: executor.Colour,
    ) -> Union[None, Awaitable[None]]:
        if response.status_code == 200:
            # Some setups are behind proxies that return some default page - for example a login - for all requests.
            # If the response is not json, we assume that this is the case and try unauthenticated access. Any auth
            # header provided by the user is unaffected.
            try:
                resp = response.json()
            except Exception:
                _Warnings.auth_cannot_parse_oidc_config(oidc_url)
                self.__make_clients(colour)
                return executor.empty(colour)

            if auth_client_secret is not None:
                if colour == "async":

                    async def execute() -> None:
                        _auth = await executor.aresult(
                            _Auth.use(
                                oidc_config=resp,
                                credentials=auth_client_secret,
                                make_mounts=lambda: self._make_mounts("async"),
                                colour=colour,
                            )
                        )
                        try:
                            self._client = await _auth.aresult(_auth.get_auth_session())
                        except HTTPError as e:
                            raise AuthenticationFailedError(
                                f"Failed to authenticate with OIDC: {repr(e)}"
                            )

                        if isinstance(auth_client_secret, AuthClientCredentials):
                            # credentials should only be saved for client credentials, otherwise use refresh token
                            self._create_background_token_refresh(_auth)
                        else:
                            self._create_background_token_refresh()

                    return execute()
                else:
                    _auth = executor.result(
                        _Auth.use(
                            oidc_config=resp,
                            credentials=auth_client_secret,
                            make_mounts=lambda: self._make_mounts("sync"),
                            colour=colour,
                        )
                    )
                    try:
                        self._client = _auth.result(_auth.get_auth_session())
                    except HTTPError as e:
                        raise AuthenticationFailedError(
                            f"Failed to authenticate with OIDC: {repr(e)}"
                        )

                    if isinstance(auth_client_secret, AuthClientCredentials):
                        # credentials should only be saved for client credentials, otherwise use refresh token
                        self._create_background_token_refresh(_auth)
                    else:
                        self._create_background_token_refresh()

            else:
                msg = f""""No login credentials provided. The weaviate instance at {self.url} requires login credentials.

                    Please check our documentation at https://weaviate.io/developers/weaviate/client-libraries/python#authentication
                    for more information about how to use authentication."""

                if is_weaviate_domain(self.url):
                    msg += """

                    You can instantiate the client with login credentials for Weaviate Cloud using

                    client = weaviate.connect_to_weaviate_cloud(
                      url=YOUR_WEAVIATE_URL,
                      auth_client_secret=wvc.init.Auth.api_key("YOUR_API_KEY")
                    )
                    """
                raise AuthenticationFailedError(msg)
        elif response.status_code == 404 and auth_client_secret is not None:
            _Warnings.auth_with_anon_weaviate()
            self.__make_clients(colour)
        else:
            self.__make_clients(colour)
        return executor.empty(colour)

    def _create_background_token_refresh(self, _auth: Optional[_Auth] = None) -> None:
        """Create a background thread that periodically refreshes access and refresh tokens.

        While the underlying library refreshes tokens, it does not have an internal cronjob that checks every
        X-seconds if a token has expired. If there is no activity for longer than the refresh tokens lifetime, it will
        expire. Therefore, refresh manually shortly before expiration time is up.
        """
        assert isinstance(self._client, (OAuth2Client, AsyncOAuth2Client))
        if "refresh_token" not in self._client.token and _auth is None:
            return

        # make an event loop sidecar thread for running async token refreshing
        event_loop = (
            _EventLoopSingleton.get_instance()
            if isinstance(self._client, AsyncOAuth2Client)
            else None
        )

        expires_in: int = self._client.token.get(
            "expires_in", 60
        )  # use 1minute as token lifetime if not supplied
        self._shutdown_background_event = Event()

        def refresh_token() -> None:
            if isinstance(self._client, AsyncOAuth2Client):
                assert event_loop is not None
                self._client.token = event_loop.run_until_complete(
                    self._client.refresh_token,
                    url=self._client.metadata["token_endpoint"],
                )
            elif isinstance(self._client, OAuth2Client):
                self._client.token = self._client.refresh_token(
                    url=self._client.metadata["token_endpoint"]
                )

        def refresh_session() -> None:
            assert _auth is not None
            if isinstance(self._client, AsyncOAuth2Client):
                assert event_loop is not None
                new_session = event_loop.run_until_complete(
                    _auth.aresult, result=_auth.get_auth_session()
                )
                self._client.token = event_loop.run_until_complete(new_session.fetch_token)
            elif isinstance(self._client, OAuth2Client):
                new_session = _auth.result(_auth.get_auth_session())
                self._client.token = new_session.fetch_token()

        def update_refresh_time() -> int:
            assert isinstance(self._client, (OAuth2Client, AsyncOAuth2Client))
            return self._client.token.get("expires_in", 60) - 30

        def periodic_refresh_token(refresh_time: int, _auth: Optional[_Auth]) -> None:
            while (
                self._shutdown_background_event is not None
                and not self._shutdown_background_event.is_set()
            ):
                # use refresh token when available
                time.sleep(max(refresh_time, 1))
                try:
                    if self._client is None:
                        continue
                    elif (
                        isinstance(self._client, (OAuth2Client, AsyncOAuth2Client))
                        and "refresh_token" in self._client.token
                    ):
                        refresh_token()
                    else:
                        # client credentials usually does not contain a refresh token => get a new token using the
                        # saved credentials
                        refresh_session()
                    refresh_time = update_refresh_time()
                except HTTPError as exc:
                    # retry again after one second, might be an unstable connection
                    refresh_time = 1
                    _Warnings.token_refresh_failed(exc)

        demon = Thread(
            target=periodic_refresh_token,
            args=(expires_in, _auth),
            daemon=True,
            name="TokenRefresh",
        )
        demon.start()

    def __get_latest_headers(self) -> Dict[str, str]:
        if "authorization" in self._headers:
            return self._headers

        auth_token = self.get_current_bearer_token()
        if auth_token == "":
            return self._headers

        # bearer token can change over time (OIDC) so we need to get the current one for each request
        copied_headers = copy(self._headers)
        copied_headers.update({"authorization": self.get_current_bearer_token()})
        self.__refresh_weaviate_embedding_service_auth_header(copied_headers)
        return copied_headers

    def __refresh_weaviate_embedding_service_auth_header(self, headers: dict[str, str]) -> None:
        if is_weaviate_domain(self._connection_params.http.host):
            # keeping for backwards compatibility for older clusters for now. On newer clusters, Embedding Service reuses Authorization header.
            headers.update({"x-weaviate-api-key": self.get_current_bearer_token()})

    def __get_timeout(
        self,
        method: Literal["DELETE", "GET", "HEAD", "PATCH", "POST", "PUT"],
        is_gql_query: bool,
    ) -> Timeout:
        """Get the timeout for the request.

        In this way, the client waits the `httpx` default of 5s when connecting to a socket (connect), writing chunks (write), and
        acquiring a connection from the pool (pool), but a custom amount as specified for reading the response (read).

        From the PoV of the user, a request is considered to be timed out if no response is received within the specified time.
        They specify the times depending on how they expect Weaviate to behave. For example, a query might take longer than an insert or vice versa
        but, in either case, the user only cares about how long it takes for a response to be received.

        https://www.python-httpx.org/advanced/timeouts/
        """
        timeout = None
        if method == "DELETE" or method == "PATCH" or method == "PUT":
            timeout = self.timeout_config.insert
        elif method == "GET" or method == "HEAD":
            timeout = self.timeout_config.query
        elif method == "POST" and is_gql_query:
            timeout = self.timeout_config.query
        elif method == "POST" and not is_gql_query:
            timeout = self.timeout_config.insert
        return Timeout(
            timeout=5.0,
            read=timeout,
            pool=self.__connection_config.session_pool_timeout,
        )

    def __handle_exceptions(self, e: Exception, error_msg: str) -> None:
        if isinstance(e, RuntimeError):
            raise WeaviateClosedClientError() from e
        if isinstance(e, ConnectError):
            raise WeaviateConnectionError(error_msg) from e
        if isinstance(e, ReadTimeout):
            raise WeaviateTimeoutError(error_msg) from e
        raise e

    def __handle_response(
        self,
        response: Response,
        error_msg: str,
        status_codes: Optional[_ExpectedStatusCodes],
    ) -> Response:
        if response.status_code == 403:
            raise InsufficientPermissionsError(response)
        if status_codes is not None and response.status_code not in status_codes.ok:
            raise UnexpectedStatusCodeError(error_msg, response)
        return response

    def _send(
        self,
        method: Literal["DELETE", "GET", "HEAD", "PATCH", "POST", "PUT"],
        *,
        url: str,
        error_msg: str,
        status_codes: Optional[_ExpectedStatusCodes],
        is_gql_query: bool = False,
        weaviate_object: Optional[JSONPayload] = None,
        params: Optional[Dict[str, Any]] = None,
        check_is_connected: bool = True,
    ) -> executor.Result[Response]:
        if check_is_connected and not self.is_connected():
            raise WeaviateClosedClientError()
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        assert self._client is not None
        request = self._client.build_request(
            method,
            url,
            json=weaviate_object,
            params=params,
            headers=self.__get_latest_headers(),
            timeout=self.__get_timeout(method, is_gql_query),
        )

        def resp(res: Response) -> Response:
            return self.__handle_response(res, error_msg, status_codes)

        def exc(e: Exception) -> None:
            self.__handle_exceptions(e, error_msg)

        return executor.execute(
            response_callback=resp,
            exception_callback=exc,
            method=self._client.send,
            request=request,
        )

    def close(self, colour: executor.Colour) -> executor.Result[None]:
        if self.embedded_db is not None:
            self.embedded_db.stop()
        if colour == "async":

            async def execute() -> None:
                if self._client is not None:
                    assert isinstance(self._client, AsyncClient)
                    await self._client.aclose()
                    self._client = None
                if self._grpc_stub is not None:
                    assert self._grpc_channel is not None
                    assert isinstance(self._grpc_channel, AsyncChannel)
                    await self._grpc_channel.close()
                    self._grpc_stub = None
                    self._grpc_channel = None
                self._connected = False

            return execute()
        if self._client is not None:
            assert isinstance(self._client, Client)
            self._client.close()
            self._client = None
        if self._grpc_stub is not None:
            assert self._grpc_channel is not None
            assert isinstance(self._grpc_channel, SyncChannel)
            self._grpc_channel.close()
            self._grpc_stub = None
            self._grpc_channel = None
        self._connected = False

    def _check_package_version(self, colour: executor.Colour) -> executor.Result[None]:
        def resp(res: Response) -> None:
            pkg_info: dict = res.json().get("info", {})
            latest_version = pkg_info.get("version", "unknown version")
            if is_weaviate_client_too_old(client_version, latest_version):
                _Warnings.weaviate_client_too_old_vs_latest(client_version, latest_version)

        try:
            if colour == "async":

                async def _execute() -> None:
                    async with AsyncClient() as client:
                        res = await client.get(PYPI_PACKAGE_URL, timeout=self.timeout_config.init)
                    return resp(res)

                return _execute()
            with Client() as client:
                res = client.get(PYPI_PACKAGE_URL, timeout=self.timeout_config.init)
            return resp(res)
        except RequestError:
            pass  # ignore any errors related to requests, it is a best-effort warning

    def delete(
        self,
        path: str,
        weaviate_object: Optional[JSONPayload] = None,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> executor.Result[Response]:
        return self._send(
            "DELETE",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    def patch(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> executor.Result[Response]:
        return self._send(
            "PATCH",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    def post(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
        is_gql_query: bool = False,
    ) -> executor.Result[Response]:
        return self._send(
            "POST",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
            is_gql_query=is_gql_query,
        )

    def put(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> executor.Result[Response]:
        return self._send(
            "PUT",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
        check_is_connected: bool = True,
    ) -> executor.Result[Response]:
        return self._send(
            "GET",
            url=self.url + self._api_version_path + path,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
            check_is_connected=check_is_connected,
        )

    def head(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> executor.Result[Response]:
        return self._send(
            "HEAD",
            url=self.url + self._api_version_path + path,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    def get_meta(self, check_is_connected: bool = True) -> executor.Result[Dict[str, str]]:
        def resp(res: Response) -> Dict[str, str]:
            data = _decode_json_response_dict(res, "Meta endpoint")
            assert data is not None
            return data

        return executor.execute(
            response_callback=resp,
            method=self.get,
            path="/meta",
            check_is_connected=check_is_connected,
        )

    def get_open_id_configuration(self) -> executor.Result[Optional[Dict[str, Any]]]:
        def resp(res: Response) -> Optional[Dict[str, Any]]:
            if res.status_code == 200:
                return _decode_json_response_dict(res, "OpenID Configuration")
            if res.status_code == 404:
                return None
            raise UnexpectedStatusCodeError("Meta endpoint", res)

        return executor.execute(
            response_callback=resp,
            method=self.get,
            path="/.well-known/openid-configuration",
        )


class ConnectionSync(_ConnectionBase):
    """Connection class used to communicate to a weaviate instance."""

    def connect(self) -> None:
        if self._connected:
            return None

        self._open_connections_rest(self._auth, "sync")

        # need this to get the version of weaviate for version checks and proper GRPC configuration
        try:
            meta = executor.result(self.get_meta(False))
            self._weaviate_version = _ServerVersion.from_string(meta["version"])
            if "grpcMaxMessageSize" in meta:
                self._grpc_max_msg_size = int(meta["grpcMaxMessageSize"])
            # Add warning later, when weaviate supported it for a while
            # else:
            #     _Warnings.grpc_max_msg_size_not_found()
        except (
            WeaviateConnectionError,
            ReadError,
            RemoteProtocolError,
            SSLZeroReturnError,  # required for async 3.8,3.9 due to ssl.SSLZeroReturnError: TLS/SSL connection has been closed (EOF) (_ssl.c:1131)
        ) as e:
            self._connected = False
            raise WeaviateStartUpError(f"Could not connect to Weaviate:{e}.") from e

        self.open_connection_grpc("sync")
        if self.embedded_db is not None:
            try:
                self.wait_for_weaviate(10)
            except WeaviateStartUpError as e:
                self.embedded_db.stop()
                self._connected = False
                raise e

        # do it after all other init checks so as not to break all the tests
        if self._weaviate_version.is_lower_than(1, 27, patch=0):
            self._connected = False
            raise WeaviateStartUpError(
                f"Weaviate version {self._weaviate_version} is not supported. Please use Weaviate version 1.27.0 or higher."
            )

        if not self._skip_init_checks:
            try:
                executor.result(self._ping_grpc("sync"))
                executor.result(self._check_package_version("sync"))
            except Exception as e:
                self._connected = False
                raise e

        self._connected = True

    def wait_for_weaviate(self, startup_period: int) -> None:
        for _i in range(startup_period):
            try:
                executor.result(
                    self.get("/.well-known/ready", check_is_connected=False)
                ).raise_for_status()
                return
            except (ConnectError, ReadError, TimeoutError, HTTPStatusError):
                time.sleep(1)

        try:
            executor.result(
                self.get("/.well-known/ready", check_is_connected=False)
            ).raise_for_status()
            return
        except (ConnectError, ReadError, TimeoutError, HTTPStatusError) as error:
            raise WeaviateStartUpError(
                f"Weaviate did not start up in {startup_period} seconds. Either the Weaviate URL {self.url} is wrong or Weaviate did not start up in the interval given in 'startup_period'."
            ) from error

    def grpc_search(self, request: search_get_pb2.SearchRequest) -> search_get_pb2.SearchReply:
        try:
            assert self.grpc_stub is not None
            res = _Retry(4).with_exponential_backoff(
                0,
                f"Searching in collection {request.collection}",
                self.grpc_stub.Search,
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.query,
            )
            return cast(search_get_pb2.SearchReply, res)
        except RpcError as e:
            error = cast(Call, e)
            if error.code() == StatusCode.PERMISSION_DENIED:
                raise InsufficientPermissionsError(error)
            raise WeaviateQueryError(str(error.details()), "GRPC search")  # pyright: ignore
        except WeaviateRetryError as e:
            raise WeaviateQueryError(str(e), "GRPC search")  # pyright: ignore

    def grpc_batch_objects(
        self,
        request: batch_pb2.BatchObjectsRequest,
        timeout: Union[int, float],
        max_retries: float,
    ) -> Dict[int, str]:
        try:
            assert self.grpc_stub is not None
            res = _Retry(max_retries).with_exponential_backoff(
                count=0,
                error="Batch objects",
                f=self.grpc_stub.BatchObjects,
                request=request,
                metadata=self.grpc_headers(),
                timeout=timeout,
            )
            res = cast(batch_pb2.BatchObjectsReply, res)

            objects: Dict[int, str] = {}
            for err in res.errors:
                objects[err.index] = err.error
            return objects
        except RpcError as e:
            error = cast(Call, e)
            if error.code() == StatusCode.PERMISSION_DENIED:
                raise InsufficientPermissionsError(error)
            raise WeaviateBatchError(str(error.details()))

    def grpc_batch_delete(
        self, request: batch_delete_pb2.BatchDeleteRequest
    ) -> batch_delete_pb2.BatchDeleteReply:
        try:
            assert self.grpc_stub is not None
            return cast(
                batch_delete_pb2.BatchDeleteReply,
                self.grpc_stub.BatchDelete(
                    request,
                    metadata=self.grpc_headers(),
                    timeout=self.timeout_config.insert,
                ),
            )
        except RpcError as e:
            error = cast(Call, e)
            if error.code() == StatusCode.PERMISSION_DENIED:
                raise InsufficientPermissionsError(error)
            raise WeaviateDeleteManyError(str(error.details()))

    def grpc_tenants_get(
        self, request: tenants_pb2.TenantsGetRequest
    ) -> tenants_pb2.TenantsGetReply:
        try:
            assert self.grpc_stub is not None
            res = _Retry().with_exponential_backoff(
                0,
                f"Get tenants for collection {request.collection}",
                self.grpc_stub.TenantsGet,
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.query,
            )
        except RpcError as e:
            error = cast(Call, e)
            if error.code() == StatusCode.PERMISSION_DENIED:
                raise InsufficientPermissionsError(error)
            raise WeaviateTenantGetError(str(error.details())) from e

        return cast(tenants_pb2.TenantsGetReply, res)

    def grpc_aggregate(
        self, request: aggregate_pb2.AggregateRequest
    ) -> aggregate_pb2.AggregateReply:
        try:
            assert self.grpc_stub is not None
            res = _Retry(4).with_exponential_backoff(
                0,
                f"Searching in collection {request.collection}",
                self.grpc_stub.Aggregate,
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.query,
            )
            return cast(aggregate_pb2.AggregateReply, res)
        except RpcError as e:
            error = cast(Call, e)
            if error.code() == StatusCode.PERMISSION_DENIED:
                raise InsufficientPermissionsError(error)
            raise WeaviateQueryError(str(e), "GRPC search")  # pyright: ignore
        except WeaviateRetryError as e:
            raise WeaviateQueryError(str(e), "GRPC search")  # pyright: ignore


class ConnectionAsync(_ConnectionBase):
    """Connection class used to communicate to a weaviate instance."""

    async def connect(self) -> None:
        if self._connected:
            return None

        await executor.aresult(self._open_connections_rest(self._auth, "async"))

        # need this to get the version of weaviate for version checks and proper GRPC configuration
        try:
            meta = await self.get_meta(False)
            self._weaviate_version = _ServerVersion.from_string(meta["version"])
            if "grpcMaxMessageSize" in meta:
                self._grpc_max_msg_size = int(meta["grpcMaxMessageSize"])
            # Add warning later, when weaviate supported it for a while
            # else:
            #     _Warnings.grpc_max_msg_size_not_found()
        except (
            WeaviateConnectionError,
            ReadError,
            RemoteProtocolError,
            SSLZeroReturnError,  # required for async 3.8,3.9 due to ssl.SSLZeroReturnError: TLS/SSL connection has been closed (EOF) (_ssl.c:1131)
        ) as e:
            self._connected = False
            raise WeaviateStartUpError(f"Could not connect to Weaviate:{e}.") from e

        self.open_connection_grpc("async")
        if self.embedded_db is not None:
            try:
                await self.wait_for_weaviate(10)
            except WeaviateStartUpError as e:
                self.embedded_db.stop()
                self._connected = False
                raise e

        # do it after all other init checks so as not to break all the tests
        if self._weaviate_version.is_lower_than(1, 27, 0):
            self._connected = False
            raise WeaviateStartUpError(
                f"Weaviate version {self._weaviate_version} is not supported. Please use Weaviate version 1.27.0 or higher."
            )

        if not self._skip_init_checks:
            try:
                await executor.aresult(self._ping_grpc("async"))
                await executor.aresult(self._check_package_version("async"))
            except Exception as e:
                self._connected = False
                raise e

        self._connected = True

    async def wait_for_weaviate(self, startup_period: int) -> None:
        for _i in range(startup_period):
            try:
                (
                    await executor.aresult(self.get("/.well-known/ready", check_is_connected=False))
                ).raise_for_status()
                return
            except (ConnectError, ReadError, TimeoutError, HTTPStatusError):
                time.sleep(1)

        try:
            (
                await executor.aresult(self.get("/.well-known/ready", check_is_connected=False))
            ).raise_for_status()
            return
        except (ConnectError, ReadError, TimeoutError, HTTPStatusError) as error:
            raise WeaviateStartUpError(
                f"Weaviate did not start up in {startup_period} seconds. Either the Weaviate URL {self.url} is wrong or Weaviate did not start up in the interval given in 'startup_period'."
            ) from error

    async def grpc_search(
        self, request: search_get_pb2.SearchRequest
    ) -> search_get_pb2.SearchReply:
        try:
            assert self.grpc_stub is not None
            res = await _Retry(4).awith_exponential_backoff(
                0,
                f"Searching in collection {request.collection}",
                self.grpc_stub.Search,
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.query,
            )
            return cast(search_get_pb2.SearchReply, res)
        except AioRpcError as e:
            if e.code().name == PERMISSION_DENIED:
                raise InsufficientPermissionsError(e)
            raise WeaviateQueryError(str(e.details()), "GRPC search")  # pyright: ignore
        except WeaviateRetryError as e:
            raise WeaviateQueryError(str(e), "GRPC search")  # pyright: ignore

    async def grpc_batch_objects(
        self,
        request: batch_pb2.BatchObjectsRequest,
        timeout: Union[int, float],
        max_retries: float,
    ) -> Dict[int, str]:
        try:
            assert self.grpc_stub is not None
            res = await _Retry(max_retries).awith_exponential_backoff(
                count=0,
                error="Batch objects",
                f=self.grpc_stub.BatchObjects,
                request=request,
                metadata=self.grpc_headers(),
                timeout=timeout,
            )
            res = cast(batch_pb2.BatchObjectsReply, res)

            objects: Dict[int, str] = {}
            for err in res.errors:
                objects[err.index] = err.error
            return objects
        except AioRpcError as e:
            if e.code().name == PERMISSION_DENIED:
                raise InsufficientPermissionsError(e)
            raise WeaviateBatchError(str(e)) from e

    async def grpc_batch_delete(
        self, request: batch_delete_pb2.BatchDeleteRequest
    ) -> batch_delete_pb2.BatchDeleteReply:
        try:
            assert self.grpc_stub is not None
            return await self.grpc_stub.BatchDelete(
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.insert,
            )
        except AioRpcError as e:
            if e.code().name == PERMISSION_DENIED:
                raise InsufficientPermissionsError(e)
            raise WeaviateDeleteManyError(str(e))

    async def grpc_tenants_get(
        self, request: tenants_pb2.TenantsGetRequest
    ) -> tenants_pb2.TenantsGetReply:
        try:
            assert self.grpc_stub is not None
            res = await _Retry().awith_exponential_backoff(
                0,
                f"Get tenants for collection {request.collection}",
                self.grpc_stub.TenantsGet,
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.query,
            )
        except AioRpcError as e:
            if e.code().name == PERMISSION_DENIED:
                raise InsufficientPermissionsError(e)
            raise WeaviateTenantGetError(str(e)) from e

        return cast(tenants_pb2.TenantsGetReply, res)

    async def grpc_aggregate(
        self, request: aggregate_pb2.AggregateRequest
    ) -> aggregate_pb2.AggregateReply:
        try:
            assert self.grpc_stub is not None
            res = await _Retry(4).awith_exponential_backoff(
                0,
                f"Searching in collection {request.collection}",
                self.grpc_stub.Aggregate,
                request,
                metadata=self.grpc_headers(),
                timeout=self.timeout_config.query,
            )
            return cast(aggregate_pb2.AggregateReply, res)
        except AioRpcError as e:
            if e.code().name == PERMISSION_DENIED:
                raise InsufficientPermissionsError(e)
            raise WeaviateQueryError(str(e), "GRPC search")  # pyright: ignore
        except WeaviateRetryError as e:
            raise WeaviateQueryError(str(e), "GRPC search")  # pyright: ignore


ConnectionV4 = ConnectionAsync
Connection = Union[ConnectionSync, ConnectionAsync]
ConnectionType = TypeVar("ConnectionType", ConnectionSync, ConnectionAsync)



from typing import Generic

from weaviate.collections.classes.types import Properties, References
from weaviate.collections.queries.near_text.query.executor import _NearTextQueryExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionAsync


@executor.wrap("async")
class _NearTextQueryAsync(
    Generic[Properties, References],
    _NearTextQueryExecutor[ConnectionAsync, Properties, References],
):
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/base.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rv1/base.proto\x12\x0bweaviate.v1\x1a\x1cgoogle/protobuf/struct.proto\"T\n\x15NumberArrayProperties\x12\x12\n\x06values\x18\x01 \x03(\x01\x42\x02\x18\x01\x12\x11\n\tprop_name\x18\x02 \x01(\t\x12\x14\n\x0cvalues_bytes\x18\x03 \x01(\x0c\"7\n\x12IntArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\x03\x12\x11\n\tprop_name\x18\x02 \x01(\t\"8\n\x13TextArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\t\x12\x11\n\tprop_name\x18\x02 \x01(\t\";\n\x16\x42ooleanArrayProperties\x12\x0e\n\x06values\x18\x01 \x03(\x08\x12\x11\n\tprop_name\x18\x02 \x01(\t\"\xf1\x03\n\x15ObjectPropertiesValue\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x43\n\x17number_array_properties\x18\x02 \x03(\x0b\x32\".weaviate.v1.NumberArrayProperties\x12=\n\x14int_array_properties\x18\x03 \x03(\x0b\x32\x1f.weaviate.v1.IntArrayProperties\x12?\n\x15text_array_properties\x18\x04 \x03(\x0b\x32 .weaviate.v1.TextArrayProperties\x12\x45\n\x18\x62oolean_array_properties\x18\x05 \x03(\x0b\x32#.weaviate.v1.BooleanArrayProperties\x12\x38\n\x11object_properties\x18\x06 \x03(\x0b\x32\x1d.weaviate.v1.ObjectProperties\x12\x43\n\x17object_array_properties\x18\x07 \x03(\x0b\x32\".weaviate.v1.ObjectArrayProperties\x12\x18\n\x10\x65mpty_list_props\x18\n \x03(\t\"^\n\x15ObjectArrayProperties\x12\x32\n\x06values\x18\x01 \x03(\x0b\x32\".weaviate.v1.ObjectPropertiesValue\x12\x11\n\tprop_name\x18\x02 \x01(\t\"X\n\x10ObjectProperties\x12\x31\n\x05value\x18\x01 \x01(\x0b\x32\".weaviate.v1.ObjectPropertiesValue\x12\x11\n\tprop_name\x18\x02 \x01(\t\"\x1b\n\tTextArray\x12\x0e\n\x06values\x18\x01 \x03(\t\"\x1a\n\x08IntArray\x12\x0e\n\x06values\x18\x01 \x03(\x03\"\x1d\n\x0bNumberArray\x12\x0e\n\x06values\x18\x01 \x03(\x01\"\x1e\n\x0c\x42ooleanArray\x12\x0e\n\x06values\x18\x01 \x03(\x08\"\xaa\x07\n\x07\x46ilters\x12/\n\x08operator\x18\x01 \x01(\x0e\x32\x1d.weaviate.v1.Filters.Operator\x12\x0e\n\x02on\x18\x02 \x03(\tB\x02\x18\x01\x12%\n\x07\x66ilters\x18\x03 \x03(\x0b\x32\x14.weaviate.v1.Filters\x12\x14\n\nvalue_text\x18\x04 \x01(\tH\x00\x12\x13\n\tvalue_int\x18\x05 \x01(\x03H\x00\x12\x17\n\rvalue_boolean\x18\x06 \x01(\x08H\x00\x12\x16\n\x0cvalue_number\x18\x07 \x01(\x01H\x00\x12\x32\n\x10value_text_array\x18\t \x01(\x0b\x32\x16.weaviate.v1.TextArrayH\x00\x12\x30\n\x0fvalue_int_array\x18\n \x01(\x0b\x32\x15.weaviate.v1.IntArrayH\x00\x12\x38\n\x13value_boolean_array\x18\x0b \x01(\x0b\x32\x19.weaviate.v1.BooleanArrayH\x00\x12\x36\n\x12value_number_array\x18\x0c \x01(\x0b\x32\x18.weaviate.v1.NumberArrayH\x00\x12\x36\n\tvalue_geo\x18\r \x01(\x0b\x32!.weaviate.v1.GeoCoordinatesFilterH\x00\x12)\n\x06target\x18\x14 \x01(\x0b\x32\x19.weaviate.v1.FilterTarget\"\x91\x03\n\x08Operator\x12\x18\n\x14OPERATOR_UNSPECIFIED\x10\x00\x12\x12\n\x0eOPERATOR_EQUAL\x10\x01\x12\x16\n\x12OPERATOR_NOT_EQUAL\x10\x02\x12\x19\n\x15OPERATOR_GREATER_THAN\x10\x03\x12\x1f\n\x1bOPERATOR_GREATER_THAN_EQUAL\x10\x04\x12\x16\n\x12OPERATOR_LESS_THAN\x10\x05\x12\x1c\n\x18OPERATOR_LESS_THAN_EQUAL\x10\x06\x12\x10\n\x0cOPERATOR_AND\x10\x07\x12\x0f\n\x0bOPERATOR_OR\x10\x08\x12\x1d\n\x19OPERATOR_WITHIN_GEO_RANGE\x10\t\x12\x11\n\rOPERATOR_LIKE\x10\n\x12\x14\n\x10OPERATOR_IS_NULL\x10\x0b\x12\x19\n\x15OPERATOR_CONTAINS_ANY\x10\x0c\x12\x19\n\x15OPERATOR_CONTAINS_ALL\x10\r\x12\x1a\n\x16OPERATOR_CONTAINS_NONE\x10\x0e\x12\x10\n\x0cOPERATOR_NOT\x10\x0f\x42\x0c\n\ntest_value\"T\n\x1b\x46ilterReferenceSingleTarget\x12\n\n\x02on\x18\x01 \x01(\t\x12)\n\x06target\x18\x02 \x01(\x0b\x32\x19.weaviate.v1.FilterTarget\"n\n\x1a\x46ilterReferenceMultiTarget\x12\n\n\x02on\x18\x01 \x01(\t\x12)\n\x06target\x18\x02 \x01(\x0b\x32\x19.weaviate.v1.FilterTarget\x12\x19\n\x11target_collection\x18\x03 \x01(\t\"\"\n\x14\x46ilterReferenceCount\x12\n\n\x02on\x18\x01 \x01(\t\"\xe4\x01\n\x0c\x46ilterTarget\x12\x12\n\x08property\x18\x01 \x01(\tH\x00\x12\x41\n\rsingle_target\x18\x02 \x01(\x0b\x32(.weaviate.v1.FilterReferenceSingleTargetH\x00\x12?\n\x0cmulti_target\x18\x03 \x01(\x0b\x32\'.weaviate.v1.FilterReferenceMultiTargetH\x00\x12\x32\n\x05\x63ount\x18\x04 \x01(\x0b\x32!.weaviate.v1.FilterReferenceCountH\x00\x42\x08\n\x06target\"M\n\x14GeoCoordinatesFilter\x12\x10\n\x08latitude\x18\x01 \x01(\x02\x12\x11\n\tlongitude\x18\x02 \x01(\x02\x12\x10\n\x08\x64istance\x18\x03 \x01(\x02\"\xd3\x01\n\x07Vectors\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\x05index\x18\x02 \x01(\x04\x42\x02\x18\x01\x12\x14\n\x0cvector_bytes\x18\x03 \x01(\x0c\x12-\n\x04type\x18\x04 \x01(\x0e\x32\x1f.weaviate.v1.Vectors.VectorType\"b\n\nVectorType\x12\x1b\n\x17VECTOR_TYPE_UNSPECIFIED\x10\x00\x12\x1b\n\x17VECTOR_TYPE_SINGLE_FP32\x10\x01\x12\x1a\n\x16VECTOR_TYPE_MULTI_FP32\x10\x02*\x89\x01\n\x10\x43onsistencyLevel\x12!\n\x1d\x43ONSISTENCY_LEVEL_UNSPECIFIED\x10\x00\x12\x19\n\x15\x43ONSISTENCY_LEVEL_ONE\x10\x01\x12\x1c\n\x18\x43ONSISTENCY_LEVEL_QUORUM\x10\x02\x12\x19\n\x15\x43ONSISTENCY_LEVEL_ALL\x10\x03\x42n\n#io.weaviate.client.grpc.protocol.v1B\x11WeaviateProtoBaseZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.base_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#io.weaviate.client.grpc.protocol.v1B\021WeaviateProtoBaseZ4github.com/weaviate/weaviate/grpc/generated;protocol'
  _NUMBERARRAYPROPERTIES.fields_by_name['values']._options = None
  _NUMBERARRAYPROPERTIES.fields_by_name['values']._serialized_options = b'\030\001'
  _FILTERS.fields_by_name['on']._options = None
  _FILTERS.fields_by_name['on']._serialized_options = b'\030\001'
  _VECTORS.fields_by_name['index']._options = None
  _VECTORS.fields_by_name['index']._serialized_options = b'\030\001'
  _globals['_CONSISTENCYLEVEL']._serialized_start=2828
  _globals['_CONSISTENCYLEVEL']._serialized_end=2965
  _globals['_NUMBERARRAYPROPERTIES']._serialized_start=60
  _globals['_NUMBERARRAYPROPERTIES']._serialized_end=144
  _globals['_INTARRAYPROPERTIES']._serialized_start=146
  _globals['_INTARRAYPROPERTIES']._serialized_end=201
  _globals['_TEXTARRAYPROPERTIES']._serialized_start=203
  _globals['_TEXTARRAYPROPERTIES']._serialized_end=259
  _globals['_BOOLEANARRAYPROPERTIES']._serialized_start=261
  _globals['_BOOLEANARRAYPROPERTIES']._serialized_end=320
  _globals['_OBJECTPROPERTIESVALUE']._serialized_start=323
  _globals['_OBJECTPROPERTIESVALUE']._serialized_end=820
  _globals['_OBJECTARRAYPROPERTIES']._serialized_start=822
  _globals['_OBJECTARRAYPROPERTIES']._serialized_end=916
  _globals['_OBJECTPROPERTIES']._serialized_start=918
  _globals['_OBJECTPROPERTIES']._serialized_end=1006
  _globals['_TEXTARRAY']._serialized_start=1008
  _globals['_TEXTARRAY']._serialized_end=1035
  _globals['_INTARRAY']._serialized_start=1037
  _globals['_INTARRAY']._serialized_end=1063
  _globals['_NUMBERARRAY']._serialized_start=1065
  _globals['_NUMBERARRAY']._serialized_end=1094
  _globals['_BOOLEANARRAY']._serialized_start=1096
  _globals['_BOOLEANARRAY']._serialized_end=1126
  _globals['_FILTERS']._serialized_start=1129
  _globals['_FILTERS']._serialized_end=2067
  _globals['_FILTERS_OPERATOR']._serialized_start=1652
  _globals['_FILTERS_OPERATOR']._serialized_end=2053
  _globals['_FILTERREFERENCESINGLETARGET']._serialized_start=2069
  _globals['_FILTERREFERENCESINGLETARGET']._serialized_end=2153
  _globals['_FILTERREFERENCEMULTITARGET']._serialized_start=2155
  _globals['_FILTERREFERENCEMULTITARGET']._serialized_end=2265
  _globals['_FILTERREFERENCECOUNT']._serialized_start=2267
  _globals['_FILTERREFERENCECOUNT']._serialized_end=2301
  _globals['_FILTERTARGET']._serialized_start=2304
  _globals['_FILTERTARGET']._serialized_end=2532
  _globals['_GEOCOORDINATESFILTER']._serialized_start=2534
  _globals['_GEOCOORDINATESFILTER']._serialized_end=2611
  _globals['_VECTORS']._serialized_start=2614
  _globals['_VECTORS']._serialized_end=2825
  _globals['_VECTORS_VECTORTYPE']._serialized_start=2727
  _globals['_VECTORS_VECTORTYPE']._serialized_end=2825
# @@protoc_insertion_point(module_scope)



from typing import Any, Generic, Literal, Optional, Type, Union, cast, overload

from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import METADATA, PROPERTIES, REFERENCES, Sorting
from weaviate.collections.classes.internal import (
    CrossReferences,
    QueryReturn,
    QueryReturnType,
    ReturnProperties,
    ReturnReferences,
    _QueryOptions,
)
from weaviate.collections.classes.types import (
    Properties,
    References,
    TProperties,
    TReferences,
)
from weaviate.collections.queries.base_executor import _BaseExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionType
from weaviate.proto.v1 import search_get_pb2
from weaviate.types import INCLUDE_VECTOR, UUID


class _FetchObjectsQueryExecutor(
    Generic[ConnectionType, Properties, References], _BaseExecutor[ConnectionType]
):
    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Literal[None] = None,
    ) -> executor.Result[QueryReturn[Properties, References]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: REFERENCES,
    ) -> executor.Result[QueryReturn[Properties, CrossReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Union[PROPERTIES, bool, None] = None,
        return_references: Type[TReferences],
    ) -> executor.Result[QueryReturn[Properties, TReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Literal[None] = None,
    ) -> executor.Result[QueryReturn[TProperties, References]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: REFERENCES,
    ) -> executor.Result[QueryReturn[TProperties, CrossReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Type[TProperties],
        return_references: Type[TReferences],
    ) -> executor.Result[QueryReturn[TProperties, TReferences]]: ...

    @overload
    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> executor.Result[QueryReturnType[Properties, References, TProperties, TReferences]]: ...

    def fetch_objects(
        self,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        after: Optional[UUID] = None,
        filters: Optional[_Filters] = None,
        sort: Optional[Sorting] = None,
        include_vector: INCLUDE_VECTOR = False,
        return_metadata: Optional[METADATA] = None,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
    ) -> executor.Result[QueryReturnType[Properties, References, TProperties, TReferences]]:
        """Retrieve the objects in this collection without any search.

        Args:
            limit: The maximum number of results to return. If not specified, the default limit specified by the server is returned.
            offset: The offset to start from. If not specified, the retrieval begins from the first object in the server.
            after: The UUID of the object to start from. If not specified, the retrieval begins from the first object in the server.
            filters: The filters to apply to the retrieval.
            sort: The sorting to apply to the retrieval.
            include_vector: Whether to include the vector in the results. If not specified, this is set to False.
            return_metadata: The metadata to return for each object, defaults to `None`.
            return_properties: The properties to return for each object.
            return_references: The references to return for each object.

        NOTE:
            - If `return_properties` is not provided then all properties are returned except for blob properties.
            - If `return_metadata` is not provided then no metadata is provided. Use MetadataQuery.full() to retrieve all metadata.
            - If `return_references` is not provided then no references are provided.

        Returns:
            A `QueryReturn` object that includes the searched objects.

        Raises:
            weaviate.exceptions.WeaviateGRPCQueryError: If the network connection to Weaviate fails.
        """

        def resp(
            res: search_get_pb2.SearchReply,
        ) -> QueryReturnType[Properties, References, TProperties, TReferences]:
            return cast(
                Any,
                self._result_to_query_return(
                    res,
                    _QueryOptions.from_input(
                        return_metadata,
                        return_properties,
                        include_vector,
                        self._references,
                        return_references,
                    ),
                ),
            )

        request = self._query.get(
            limit=limit,
            offset=offset,
            after=after,
            filters=filters,
            sort=sort,
            return_metadata=self._parse_return_metadata(return_metadata, include_vector),
            return_properties=self._parse_return_properties(return_properties),
            return_references=self._parse_return_references(cast(Any, return_references)),
        )
        return executor.execute(
            response_callback=resp,
            method=self._connection.grpc_search,
            request=request,
        )



from . import aggregate, backup, batch, cluster, config, data, query, replication, tenants, users

__all__ = [
    "aggregate",
    "backup",
    "batch",
    "cluster",
    "config",
    "data",
    "query",
    "replication",
    "tenants",
    "users",
]



from enum import Enum
from typing import Any, Dict, List, Optional, cast

from pydantic import BaseModel, Field

from weaviate.backup.backup_location import _BackupLocationConfig

STORAGE_NAMES = {
    "filesystem",
    "s3",
    "gcs",
    "azure",
}


class BackupCompressionLevel(str, Enum):
    """Which compression level should be used to compress the backup."""

    DEFAULT = "DefaultCompression"
    BEST_SPEED = "BestSpeed"
    BEST_COMPRESSION = "BestCompression"


class BackupStorage(str, Enum):
    """Which backend should be used to write the backup to."""

    FILESYSTEM = "filesystem"
    S3 = "s3"
    GCS = "gcs"
    AZURE = "azure"


class BackupStatus(str, Enum):
    """The status of a backup."""

    STARTED = "STARTED"
    TRANSFERRING = "TRANSFERRING"
    TRANSFERRED = "TRANSFERRED"
    SUCCESS = "SUCCESS"
    FAILED = "FAILED"
    CANCELED = "CANCELED"


class _BackupConfigBase(BaseModel):
    CPUPercentage: Optional[int] = Field(default=None, alias="cpu_percentage")

    def _to_dict(self) -> Dict[str, Any]:
        ret = cast(dict, self.model_dump(exclude_none=True))

        for key, val in ret.items():
            if isinstance(val, _BackupLocationConfig):
                ret[key] = val._to_dict()

        return ret


class BackupConfigCreate(_BackupConfigBase):
    """Options to configure the backup when creating a backup."""

    ChunkSize: Optional[int] = Field(default=None, alias="chunk_size")
    CompressionLevel: Optional[BackupCompressionLevel] = Field(
        default=None, alias="compression_level"
    )


class BackupConfigRestore(_BackupConfigBase):
    """Options to configure the backup when restoring a backup."""


class BackupStatusReturn(BaseModel):
    """Return type of the backup status methods."""

    error: Optional[str] = Field(default=None)
    status: BackupStatus
    path: str
    backup_id: str = Field(alias="id")


class BackupReturn(BackupStatusReturn):
    """Return type of the backup creation and restore methods."""

    collections: List[str] = Field(default_factory=list, alias="classes")


class BackupListReturn(BaseModel):
    """Return type of the backup list method."""

    collections: List[str] = Field(default_factory=list, alias="classes")
    status: BackupStatus
    backup_id: str = Field(alias="id")



import pytest

from weaviate.config import ConnectionConfig
from weaviate.connect import ConnectionParams, ConnectionV4


@pytest.fixture
def connection() -> ConnectionV4:
    con = ConnectionV4(
        ConnectionParams.from_url("http://localhost:8080", 50051),
        None,
        (10, 60),
        None,
        True,
        None,
        ConnectionConfig(),
        None,
    )
    con._Connection__connected = True
    return con



from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, Optional, overload

from pydantic import Field
from typing_extensions import deprecated

from weaviate.collections.classes.config_base import (
    _ConfigCreateModel,
    _ConfigUpdateModel,
    _QuantizerConfigCreate,
    _QuantizerConfigUpdate,
)
from weaviate.collections.classes.config_vectorizers import VectorDistances
from weaviate.str_enum import BaseEnum
from weaviate.warnings import _Warnings


class VectorFilterStrategy(str, Enum):
    """Set the strategy when doing a filtered HNSW search.

    Attributes:
        SWEEPING: Do normal ANN search and skip nodes.
        ACORN: Multi-hop search to find new candidates matching the filter.
    """

    SWEEPING = "sweeping"
    ACORN = "acorn"


class VectorIndexType(str, Enum):
    """The available vector index types in Weaviate.

    Attributes:
        HNSW: Hierarchical Navigable Small World (HNSW) index.
        FLAT: Flat index.
    """

    HNSW = "hnsw"
    FLAT = "flat"
    DYNAMIC = "dynamic"


class _MultiVectorConfigCreateBase(_ConfigCreateModel):
    enabled: bool = Field(default=True)


class _MultiVectorEncodingConfigCreate(_MultiVectorConfigCreateBase):
    enabled: bool = Field(default=True)

    @staticmethod
    @abstractmethod
    def encoding_name() -> str: ...


class _MuveraConfigCreate(_MultiVectorEncodingConfigCreate):
    ksim: Optional[int]
    dprojections: Optional[int]
    repetitions: Optional[int]

    @staticmethod
    def encoding_name() -> str:
        return "muvera"


class _MultiVectorConfigCreate(_MultiVectorConfigCreateBase):
    encoding: Optional[_MultiVectorEncodingConfigCreate] = Field(exclude=True)
    aggregation: Optional[str]


class _VectorIndexConfigCreate(_ConfigCreateModel):
    distance: Optional[VectorDistances]
    multivector: Optional[_MultiVectorConfigCreate]
    quantizer: Optional[_QuantizerConfigCreate] = Field(exclude=True)

    @staticmethod
    @abstractmethod
    def vector_index_type() -> VectorIndexType: ...

    def _to_dict(self) -> Dict[str, Any]:
        ret_dict = super()._to_dict()
        if self.quantizer is not None:
            if isinstance(self.quantizer, _UncompressedConfigCreate):
                ret_dict[self.quantizer.quantizer_name()] = True
            else:
                ret_dict[self.quantizer.quantizer_name()] = self.quantizer._to_dict()
        if self.distance is not None:
            ret_dict["distance"] = str(self.distance.value)
        if self.multivector is not None and self.multivector.encoding is not None:
            ret_dict["multivector"][self.multivector.encoding.encoding_name()] = (
                self.multivector.encoding._to_dict()
            )

        return ret_dict


class _VectorIndexConfigUpdate(_ConfigUpdateModel):
    quantizer: Optional[_QuantizerConfigUpdate] = Field(exclude=True)

    @staticmethod
    @abstractmethod
    def vector_index_type() -> VectorIndexType: ...


class _VectorIndexConfigSkipCreate(_VectorIndexConfigCreate):
    skip: bool = True

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.HNSW


class _VectorIndexConfigHNSWCreate(_VectorIndexConfigCreate):
    cleanupIntervalSeconds: Optional[int]
    dynamicEfMin: Optional[int]
    dynamicEfMax: Optional[int]
    dynamicEfFactor: Optional[int]
    efConstruction: Optional[int]
    ef: Optional[int]
    filterStrategy: Optional[VectorFilterStrategy]
    flatSearchCutoff: Optional[int]
    maxConnections: Optional[int]
    vectorCacheMaxObjects: Optional[int]

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.HNSW


class _VectorIndexConfigFlatCreate(_VectorIndexConfigCreate):
    vectorCacheMaxObjects: Optional[int]

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.FLAT


class _VectorIndexConfigHNSWUpdate(_VectorIndexConfigUpdate):
    dynamicEfMin: Optional[int]
    dynamicEfMax: Optional[int]
    dynamicEfFactor: Optional[int]
    ef: Optional[int]
    filterStrategy: Optional[VectorFilterStrategy]
    flatSearchCutoff: Optional[int]
    vectorCacheMaxObjects: Optional[int]

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.HNSW


class _VectorIndexConfigFlatUpdate(_VectorIndexConfigUpdate):
    vectorCacheMaxObjects: Optional[int]

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.FLAT


class _VectorIndexConfigDynamicCreate(_VectorIndexConfigCreate):
    threshold: Optional[int]
    hnsw: Optional[_VectorIndexConfigHNSWCreate]
    flat: Optional[_VectorIndexConfigFlatCreate]

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.DYNAMIC

    def _to_dict(self) -> dict:
        ret_dict = super()._to_dict()
        if self.hnsw is not None:
            ret_dict["hnsw"] = self.hnsw._to_dict()
        if self.flat is not None:
            ret_dict["flat"] = self.flat._to_dict()
        if self.threshold is not None:
            ret_dict["threshold"] = self.threshold

        return ret_dict


class _VectorIndexConfigDynamicUpdate(_VectorIndexConfigUpdate):
    threshold: Optional[int]
    hnsw: Optional[_VectorIndexConfigHNSWUpdate]
    flat: Optional[_VectorIndexConfigFlatUpdate]

    @staticmethod
    def vector_index_type() -> VectorIndexType:
        return VectorIndexType.DYNAMIC


class PQEncoderType(str, BaseEnum):
    """Type of the PQ encoder.

    Attributes:
        KMEANS: K-means encoder.
        TILE: Tile encoder.
    """

    KMEANS = "kmeans"
    TILE = "tile"


class PQEncoderDistribution(str, BaseEnum):
    """Distribution of the PQ encoder.

    Attributes:
        LOG_NORMAL: Log-normal distribution.
        NORMAL: Normal distribution.
    """

    LOG_NORMAL = "log-normal"
    NORMAL = "normal"


class MultiVectorAggregation(str, BaseEnum):
    """Aggregation type to use for multivector indices.

    Attributes:
        MAX_SIM: Maximum similarity.
    """

    MAX_SIM = "maxSim"


class _PQEncoderConfigCreate(_ConfigCreateModel):
    type_: Optional[PQEncoderType] = Field(serialization_alias="type")
    distribution: Optional[PQEncoderDistribution]


class _PQEncoderConfigUpdate(_ConfigUpdateModel):
    type_: Optional[PQEncoderType]
    distribution: Optional[PQEncoderDistribution]

    def merge_with_existing(self, schema: Dict[str, Any]) -> Dict[str, Any]:
        """Must be done manually since Pydantic does not work well with type and type_.

        Errors shadowing type occur if we want to use type as a field name.
        """
        if self.type_ is not None:
            schema["type"] = str(self.type_.value)
        if self.distribution is not None:
            schema["distribution"] = str(self.distribution.value)
        return schema


class _PQConfigCreate(_QuantizerConfigCreate):
    bitCompression: Optional[bool] = Field(default=None)
    centroids: Optional[int]
    encoder: _PQEncoderConfigCreate
    segments: Optional[int]
    trainingLimit: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "pq"


class _BQConfigCreate(_QuantizerConfigCreate):
    cache: Optional[bool]
    rescoreLimit: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "bq"


class _SQConfigCreate(_QuantizerConfigCreate):
    cache: Optional[bool]
    rescoreLimit: Optional[int]
    trainingLimit: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "sq"


class _RQConfigCreate(_QuantizerConfigCreate):
    bits: Optional[int]
    rescoreLimit: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "rq"


class _UncompressedConfigCreate(_QuantizerConfigCreate):
    @staticmethod
    def quantizer_name() -> str:
        return "skipDefaultQuantization"


class _PQConfigUpdate(_QuantizerConfigUpdate):
    bitCompression: Optional[bool] = Field(default=None)
    centroids: Optional[int]
    enabled: Optional[bool]
    segments: Optional[int]
    trainingLimit: Optional[int]
    encoder: Optional[_PQEncoderConfigUpdate]

    @staticmethod
    def quantizer_name() -> str:
        return "pq"


class _BQConfigUpdate(_QuantizerConfigUpdate):
    enabled: Optional[bool]
    rescoreLimit: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "bq"


class _RQConfigUpdate(_QuantizerConfigUpdate):
    enabled: Optional[bool]
    rescoreLimit: Optional[int]
    bits: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "rq"


class _SQConfigUpdate(_QuantizerConfigUpdate):
    enabled: Optional[bool]
    rescoreLimit: Optional[int]
    trainingLimit: Optional[int]

    @staticmethod
    def quantizer_name() -> str:
        return "sq"


class _VectorIndexMultivectorEncoding:
    @staticmethod
    def muvera(
        ksim: Optional[int] = None,
        dprojections: Optional[int] = None,
        repetitions: Optional[int] = None,
    ) -> _MultiVectorEncodingConfigCreate:
        return _MuveraConfigCreate(
            enabled=True,
            ksim=ksim,
            dprojections=dprojections,
            repetitions=repetitions,
        )


class _VectorIndexMultiVector:
    Encoding = _VectorIndexMultivectorEncoding

    @deprecated(
        'Using the "encoding" argument is deprecated. Instead, specify it at the top-level when creating your `vector_config`'
    )
    @overload
    @staticmethod
    def multi_vector(
        encoding: _MultiVectorEncodingConfigCreate,
        aggregation: Optional[MultiVectorAggregation] = None,
    ) -> _MultiVectorConfigCreate: ...

    @overload
    @staticmethod
    def multi_vector(
        encoding: Optional[_MultiVectorEncodingConfigCreate] = None,
        aggregation: Optional[MultiVectorAggregation] = None,
    ) -> _MultiVectorConfigCreate: ...

    @staticmethod
    def multi_vector(
        encoding: Optional[_MultiVectorEncodingConfigCreate] = None,
        aggregation: Optional[MultiVectorAggregation] = None,
    ) -> _MultiVectorConfigCreate:
        if encoding is not None:
            _Warnings.encoding_in_multi_vector_config()
        return _MultiVectorConfigCreate(
            encoding=encoding if encoding is not None else None,
            aggregation=aggregation.value if aggregation is not None else None,
        )


class _VectorIndexQuantizer:
    @staticmethod
    def pq(
        bit_compression: Optional[bool] = None,
        centroids: Optional[int] = None,
        encoder_distribution: Optional[PQEncoderDistribution] = None,
        encoder_type: Optional[PQEncoderType] = None,
        segments: Optional[int] = None,
        training_limit: Optional[int] = None,
    ) -> _PQConfigCreate:
        """Create a `_PQConfigCreate` object to be used when defining the product quantization (PQ) configuration of Weaviate.

        Use this method when defining the `quantizer` argument in the `vector_index` configuration.

        Args:
            See [the docs](https://weaviate.io/developers/weaviate/concepts/vector-index#hnsw-with-compression) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        if bit_compression is not None:
            _Warnings.bit_compression_in_pq_config()
        return _PQConfigCreate(
            centroids=centroids,
            segments=segments,
            trainingLimit=training_limit,
            encoder=_PQEncoderConfigCreate(type_=encoder_type, distribution=encoder_distribution),
        )

    @staticmethod
    def bq(
        cache: Optional[bool] = None,
        rescore_limit: Optional[int] = None,
    ) -> _BQConfigCreate:
        """Create a `_BQConfigCreate` object to be used when defining the binary quantization (BQ) configuration of Weaviate.

        Use this method when defining the `quantizer` argument in the `vector_index` configuration. Note that the arguments have no effect for HNSW.

        Args:
            See [the docs](https://weaviate.io/developers/weaviate/concepts/vector-index#binary-quantization) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _BQConfigCreate(
            cache=cache,
            rescoreLimit=rescore_limit,
        )

    @staticmethod
    def sq(
        cache: Optional[bool] = None,
        rescore_limit: Optional[int] = None,
        training_limit: Optional[int] = None,
    ) -> _SQConfigCreate:
        """Create a `_SQConfigCreate` object to be used when defining the scalar quantization (SQ) configuration of Weaviate.

        Use this method when defining the `quantizer` argument in the `vector_index` configuration. Note that the arguments have no effect for HNSW.

        Args:
            See [the docs](https://weaviate.io/developers/weaviate/concepts/vector-index#binary-quantization) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _SQConfigCreate(
            cache=cache,
            rescoreLimit=rescore_limit,
            trainingLimit=training_limit,
        )

    @staticmethod
    def rq(
        bits: Optional[int] = None,
        rescore_limit: Optional[int] = None,
    ) -> _RQConfigCreate:
        """Create a `_RQConfigCreate` object to be used when defining the Rotational quantization (RQ) configuration of Weaviate.

        Use this method when defining the `quantizer` argument in the `vector_index` configuration. Note that the arguments have no effect for HNSW.

        Arguments:
            See [the docs](https://weaviate.io/developers/weaviate/concepts/vector-index) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _RQConfigCreate(
            bits=bits,
            rescoreLimit=rescore_limit,
        )

    @staticmethod
    def none() -> _UncompressedConfigCreate:
        """Create a a vector index without compression."""
        return _UncompressedConfigCreate()


class _VectorIndex:
    MultiVector = _VectorIndexMultiVector
    Quantizer = _VectorIndexQuantizer

    @staticmethod
    def none() -> _VectorIndexConfigSkipCreate:
        """Create a `_VectorIndexConfigSkipCreate` object to be used when configuring Weaviate to not index your vectors.

        Use this method when defining the `vector_index_config` argument in `collections.create()`.
        """
        return _VectorIndexConfigSkipCreate(
            distance=None,
            quantizer=None,
            multivector=None,
        )

    @overload
    @staticmethod
    @deprecated(
        'Using the "multi_vector" argument is deprecated. Instead, specify it at the top-level in `multi_vector_index_config` when creating your `vector_config` with `MultiVectors.module()`'
    )
    def hnsw(
        cleanup_interval_seconds: Optional[int] = None,
        distance_metric: Optional[VectorDistances] = None,
        dynamic_ef_factor: Optional[int] = None,
        dynamic_ef_max: Optional[int] = None,
        dynamic_ef_min: Optional[int] = None,
        ef: Optional[int] = None,
        ef_construction: Optional[int] = None,
        filter_strategy: Optional[VectorFilterStrategy] = None,
        flat_search_cutoff: Optional[int] = None,
        max_connections: Optional[int] = None,
        vector_cache_max_objects: Optional[int] = None,
        *,
        quantizer: Optional[_QuantizerConfigCreate] = None,
        multi_vector: _MultiVectorConfigCreate,
    ) -> _VectorIndexConfigHNSWCreate: ...

    @overload
    @staticmethod
    def hnsw(
        cleanup_interval_seconds: Optional[int] = None,
        distance_metric: Optional[VectorDistances] = None,
        dynamic_ef_factor: Optional[int] = None,
        dynamic_ef_max: Optional[int] = None,
        dynamic_ef_min: Optional[int] = None,
        ef: Optional[int] = None,
        ef_construction: Optional[int] = None,
        filter_strategy: Optional[VectorFilterStrategy] = None,
        flat_search_cutoff: Optional[int] = None,
        max_connections: Optional[int] = None,
        vector_cache_max_objects: Optional[int] = None,
        quantizer: Optional[_QuantizerConfigCreate] = None,
        multi_vector: Optional[_MultiVectorConfigCreate] = None,
    ) -> _VectorIndexConfigHNSWCreate: ...

    @staticmethod
    def hnsw(
        cleanup_interval_seconds: Optional[int] = None,
        distance_metric: Optional[VectorDistances] = None,
        dynamic_ef_factor: Optional[int] = None,
        dynamic_ef_max: Optional[int] = None,
        dynamic_ef_min: Optional[int] = None,
        ef: Optional[int] = None,
        ef_construction: Optional[int] = None,
        filter_strategy: Optional[VectorFilterStrategy] = None,
        flat_search_cutoff: Optional[int] = None,
        max_connections: Optional[int] = None,
        vector_cache_max_objects: Optional[int] = None,
        quantizer: Optional[_QuantizerConfigCreate] = None,
        multi_vector: Optional[_MultiVectorConfigCreate] = None,
    ) -> _VectorIndexConfigHNSWCreate:
        """Create a `_VectorIndexConfigHNSWCreate` object to be used when defining the HNSW vector index configuration of Weaviate.

        Use this method when defining the `vector_index_config` argument in `collections.create()`.

        Args:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#how-to-configure-hnsw) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        if multi_vector is not None:
            _Warnings.multi_vector_in_hnsw_config()
        return _VectorIndexConfigHNSWCreate(
            cleanupIntervalSeconds=cleanup_interval_seconds,
            distance=distance_metric,
            dynamicEfMin=dynamic_ef_min,
            dynamicEfMax=dynamic_ef_max,
            dynamicEfFactor=dynamic_ef_factor,
            efConstruction=ef_construction,
            ef=ef,
            filterStrategy=filter_strategy,
            flatSearchCutoff=flat_search_cutoff,
            maxConnections=max_connections,
            vectorCacheMaxObjects=vector_cache_max_objects,
            quantizer=quantizer,
            multivector=multi_vector,
        )

    @staticmethod
    def flat(
        distance_metric: Optional[VectorDistances] = None,
        vector_cache_max_objects: Optional[int] = None,
        quantizer: Optional[_QuantizerConfigCreate] = None,
    ) -> _VectorIndexConfigFlatCreate:
        """Create a `_VectorIndexConfigFlatCreate` object to be used when defining the FLAT vector index configuration of Weaviate.

        Use this method when defining the `vector_index_config` argument in `collections.create()`.

        Args:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#how-to-configure-hnsw) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _VectorIndexConfigFlatCreate(
            distance=distance_metric,
            vectorCacheMaxObjects=vector_cache_max_objects,
            quantizer=quantizer,
            multivector=None,
        )

    @staticmethod
    def dynamic(
        distance_metric: Optional[VectorDistances] = None,
        threshold: Optional[int] = None,
        hnsw: Optional[_VectorIndexConfigHNSWCreate] = None,
        flat: Optional[_VectorIndexConfigFlatCreate] = None,
    ) -> _VectorIndexConfigDynamicCreate:
        """Create a `_VectorIndexConfigDynamicCreate` object to be used when defining the DYNAMIC vector index configuration of Weaviate.

        Use this method when defining the `vector_index_config` argument in `collections.create()`.

        Args:
            See [the docs](https://weaviate.io/developers/weaviate/configuration/indexes#how-to-configure-hnsw) for a more detailed view!
        """  # noqa: D417 (missing argument descriptions in the docstring)
        return _VectorIndexConfigDynamicCreate(
            distance=distance_metric,
            threshold=threshold,
            hnsw=hnsw,
            flat=flat,
            quantizer=None,
            multivector=None,
        )



from typing import List, Literal, Optional, Union, overload

from weaviate.collections.classes.aggregate import (
    AggregateGroupByReturn,
    AggregateReturn,
    GroupByAggregate,
    PropertiesMetrics,
)
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import BM25OperatorOptions
from weaviate.connect.v4 import ConnectionSync
from weaviate.types import NUMBER

from .executor import _HybridExecutor

class _Hybrid(_HybridExecutor[ConnectionSync]):
    @overload
    def hybrid(
        self,
        query: Optional[str],
        *,
        alpha: NUMBER = 0.7,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        object_limit: Optional[int] = None,
        bm25_operator: Optional[BM25OperatorOptions] = None,
        filters: Optional[_Filters] = None,
        group_by: Literal[None] = None,
        target_vector: Optional[str] = None,
        max_vector_distance: Optional[float] = None,
        total_count: bool = True,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> AggregateReturn: ...
    @overload
    def hybrid(
        self,
        query: Optional[str],
        *,
        alpha: NUMBER = 0.7,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        object_limit: Optional[int] = None,
        bm25_operator: Optional[BM25OperatorOptions] = None,
        filters: Optional[_Filters] = None,
        group_by: Union[str, GroupByAggregate],
        target_vector: Optional[str] = None,
        max_vector_distance: Optional[float] = None,
        total_count: bool = True,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> AggregateGroupByReturn: ...
    @overload
    def hybrid(
        self,
        query: Optional[str],
        *,
        alpha: NUMBER = 0.7,
        vector: Optional[List[float]] = None,
        query_properties: Optional[List[str]] = None,
        object_limit: Optional[int] = None,
        bm25_operator: Optional[BM25OperatorOptions] = None,
        filters: Optional[_Filters] = None,
        group_by: Optional[Union[str, GroupByAggregate]] = None,
        target_vector: Optional[str] = None,
        max_vector_distance: Optional[float] = None,
        total_count: bool = True,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> Union[AggregateReturn, AggregateGroupByReturn]: ...



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import contextlib
import os
import sys
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as version_func

# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.insert(0, os.path.abspath(".."))

# weaviate-agents-python-client as a sub-library (cloned into docs/)
sys.path.insert(0, os.path.abspath("weaviate-agents-python-client"))


@contextlib.contextmanager
def chdir(directory):
    curdir = os.curdir
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(curdir)


try:
    version = version_func("weaviate")
except PackageNotFoundError:
    # The project is not installed in readthedocs environment. Read the version with setuptools_scm.
    import setuptools_scm

    with chdir(".."):
        release = setuptools_scm.get_version()


# -- Project information -----------------------------------------------------

project = "Weaviate Python Client"
copyright = "2021-2025, Weaviate"
author = "Weaviate"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.autodoc_pydantic",
]

# Autodoc settings for pydantic
autodoc_pydantic_model_show_json = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_model_show_validator_summary = False
autodoc_pydantic_model_show_validator_members = False
autodoc_pydantic_model_show_field_summary = False
autodoc_pydantic_model_undoc_members = False
autodoc_pydantic_model_members = False

autodoc_typehints = "description"
autodoc_member_order = "bysource"
autodoc_dataclass_fields = False

# Make sure the target is unique
autosectionlabel_prefix_document = True

autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "README.rst"]

suppress_warnings = [
    "docutils",
    "autodoc",
    "autosectionlabel",
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_theme_options = {"navigation_depth": 10}  # Increase this to match `toctree`

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

import re


def convert_markdown_links(lines):
    """Convert Markdown-style [text](url) links to reST-style `text <url>`_ links."""
    md_link_pattern = re.compile(r"\[([^\]]+)\]\((http[^\)]+)\)")
    return [md_link_pattern.sub(r"`\1 <\2>`_", line) for line in lines]


def replace_client_parent_docstring_to_match_child(what, obj, lines):
    """Replace the parent class docstring with the child class docstring."""
    if what != "class" or not obj.__name__ in ("WeaviateClient", "WeaviateAsyncClient"):
        return lines

    # Replace Parent class's "WeaviateClient/WeaviateClientAsync" with the child class's name
    text = "\n".join(lines)
    text = text.replace("WeaviateClient/WeaviateClientAsync", obj.__name__)

    # Make the connect_to function references into links to actual functions
    pattern = re.compile(r"\b(weaviate\.connect_to_[a-zA-Z_][a-zA-Z0-9_]*)\b")
    text = pattern.sub(r":func:`\1`", text)

    # Rename the connect_to function to use_async_with for the WeaviateAsyncClient
    if obj.__name__ == "WeaviateAsyncClient":
        text = text.replace("connect_to", "use_async_with")

    return text.split("\n")


def shorthand_weaviate_exceptions_display(lines):
    """Replace weaviate.exceptions.* with ~weaviate.exceptions.* to make it a shorthand."""
    pattern = re.compile(r"\b(weaviate\.exceptions\.[a-zA-Z_][a-zA-Z0-9_]*)\b")
    return [pattern.sub(r"~\1", line) for line in lines]


def autodoc_process_docstring(app, what, name, obj, options, lines):
    """Apply the conversion to all docstrings."""
    lines[:] = convert_markdown_links(lines)
    lines[:] = replace_client_parent_docstring_to_match_child(what, obj, lines)
    lines[:] = shorthand_weaviate_exceptions_display(lines)


def setup(app):
    app.add_css_file("custom.css")
    app.connect("autodoc-process-docstring", autodoc_process_docstring)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: v1/batch.proto
# Protobuf Python Version: 6.30.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    0,
    '',
    'v1/batch.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from weaviate.proto.v1.v6300.v1 import base_pb2 as v1_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ev1/batch.proto\x12\x0bweaviate.v1\x1a\x1cgoogle/protobuf/struct.proto\x1a\rv1/base.proto\"\x95\x01\n\x13\x42\x61tchObjectsRequest\x12)\n\x07objects\x18\x01 \x03(\x0b\x32\x18.weaviate.v1.BatchObject\x12=\n\x11\x63onsistency_level\x18\x02 \x01(\x0e\x32\x1d.weaviate.v1.ConsistencyLevelH\x00\x88\x01\x01\x42\x14\n\x12_consistency_level\"\x9e\x01\n\x16\x42\x61tchReferencesRequest\x12/\n\nreferences\x18\x01 \x03(\x0b\x32\x1b.weaviate.v1.BatchReference\x12=\n\x11\x63onsistency_level\x18\x02 \x01(\x0e\x32\x1d.weaviate.v1.ConsistencyLevelH\x00\x88\x01\x01\x42\x14\n\x12_consistency_level\"\xd6\x02\n\x10\x42\x61tchSendRequest\x12\x11\n\tstream_id\x18\x01 \x01(\t\x12\x38\n\x07objects\x18\x02 \x01(\x0b\x32%.weaviate.v1.BatchSendRequest.ObjectsH\x00\x12>\n\nreferences\x18\x03 \x01(\x0b\x32(.weaviate.v1.BatchSendRequest.ReferencesH\x00\x12\x32\n\x04stop\x18\x04 \x01(\x0b\x32\".weaviate.v1.BatchSendRequest.StopH\x00\x1a\x06\n\x04Stop\x1a\x33\n\x07Objects\x12(\n\x06values\x18\x01 \x03(\x0b\x32\x18.weaviate.v1.BatchObject\x1a\x39\n\nReferences\x12+\n\x06values\x18\x01 \x03(\x0b\x32\x1b.weaviate.v1.BatchReferenceB\t\n\x07message\"B\n\x0e\x42\x61tchSendReply\x12\x17\n\x0fnext_batch_size\x18\x01 \x01(\x05\x12\x17\n\x0f\x62\x61\x63koff_seconds\x18\x02 \x01(\x02\"\xc7\x01\n\x12\x42\x61tchStreamRequest\x12=\n\x11\x63onsistency_level\x18\x01 \x01(\x0e\x32\x1d.weaviate.v1.ConsistencyLevelH\x00\x88\x01\x01\x12\x19\n\x0cobject_index\x18\x02 \x01(\x05H\x01\x88\x01\x01\x12\x1c\n\x0freference_index\x18\x03 \x01(\x05H\x02\x88\x01\x01\x42\x14\n\x12_consistency_levelB\x0f\n\r_object_indexB\x12\n\x10_reference_index\"\xf0\x03\n\x12\x42\x61tchStreamMessage\x12\x11\n\tstream_id\x18\x01 \x01(\t\x12\x36\n\x05\x65rror\x18\x02 \x01(\x0b\x32%.weaviate.v1.BatchStreamMessage.ErrorH\x00\x12\x36\n\x05start\x18\x03 \x01(\x0b\x32%.weaviate.v1.BatchStreamMessage.StartH\x00\x12\x34\n\x04stop\x18\x04 \x01(\x0b\x32$.weaviate.v1.BatchStreamMessage.StopH\x00\x12<\n\x08shutdown\x18\x05 \x01(\x0b\x32(.weaviate.v1.BatchStreamMessage.ShutdownH\x00\x12\x45\n\rshutting_down\x18\x06 \x01(\x0b\x32,.weaviate.v1.BatchStreamMessage.ShuttingDownH\x00\x1a\x07\n\x05Start\x1a\x06\n\x04Stop\x1a\n\n\x08Shutdown\x1a\x0e\n\x0cShuttingDown\x1a\x64\n\x05\x45rror\x12\r\n\x05\x65rror\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x05\x12\x14\n\x0cis_retriable\x18\x03 \x01(\x08\x12\x11\n\tis_object\x18\x04 \x01(\x08\x12\x14\n\x0cis_reference\x18\x05 \x01(\x08\x42\t\n\x07message\"\xde\x07\n\x0b\x42\x61tchObject\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x12\n\x06vector\x18\x02 \x03(\x02\x42\x02\x18\x01\x12\x37\n\nproperties\x18\x03 \x01(\x0b\x32#.weaviate.v1.BatchObject.Properties\x12\x12\n\ncollection\x18\x04 \x01(\t\x12\x0e\n\x06tenant\x18\x05 \x01(\t\x12\x14\n\x0cvector_bytes\x18\x06 \x01(\x0c\x12%\n\x07vectors\x18\x17 \x03(\x0b\x32\x14.weaviate.v1.Vectors\x1a\x84\x05\n\nProperties\x12\x33\n\x12non_ref_properties\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12N\n\x17single_target_ref_props\x18\x02 \x03(\x0b\x32-.weaviate.v1.BatchObject.SingleTargetRefProps\x12L\n\x16multi_target_ref_props\x18\x03 \x03(\x0b\x32,.weaviate.v1.BatchObject.MultiTargetRefProps\x12\x43\n\x17number_array_properties\x18\x04 \x03(\x0b\x32\".weaviate.v1.NumberArrayProperties\x12=\n\x14int_array_properties\x18\x05 \x03(\x0b\x32\x1f.weaviate.v1.IntArrayProperties\x12?\n\x15text_array_properties\x18\x06 \x03(\x0b\x32 .weaviate.v1.TextArrayProperties\x12\x45\n\x18\x62oolean_array_properties\x18\x07 \x03(\x0b\x32#.weaviate.v1.BooleanArrayProperties\x12\x38\n\x11object_properties\x18\x08 \x03(\x0b\x32\x1d.weaviate.v1.ObjectProperties\x12\x43\n\x17object_array_properties\x18\t \x03(\x0b\x32\".weaviate.v1.ObjectArrayProperties\x12\x18\n\x10\x65mpty_list_props\x18\n \x03(\t\x1a\x38\n\x14SingleTargetRefProps\x12\r\n\x05uuids\x18\x01 \x03(\t\x12\x11\n\tprop_name\x18\x02 \x01(\t\x1aR\n\x13MultiTargetRefProps\x12\r\n\x05uuids\x18\x01 \x03(\t\x12\x11\n\tprop_name\x18\x02 \x01(\t\x12\x19\n\x11target_collection\x18\x03 \x01(\t\"\x99\x01\n\x0e\x42\x61tchReference\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x17\n\x0f\x66rom_collection\x18\x02 \x01(\t\x12\x11\n\tfrom_uuid\x18\x03 \x01(\t\x12\x1a\n\rto_collection\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x0f\n\x07to_uuid\x18\x05 \x01(\t\x12\x0e\n\x06tenant\x18\x06 \x01(\tB\x10\n\x0e_to_collection\"\x88\x01\n\x11\x42\x61tchObjectsReply\x12\x0c\n\x04took\x18\x01 \x01(\x02\x12\x39\n\x06\x65rrors\x18\x02 \x03(\x0b\x32).weaviate.v1.BatchObjectsReply.BatchError\x1a*\n\nBatchError\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\x8e\x01\n\x14\x42\x61tchReferencesReply\x12\x0c\n\x04took\x18\x01 \x01(\x02\x12<\n\x06\x65rrors\x18\x02 \x03(\x0b\x32,.weaviate.v1.BatchReferencesReply.BatchError\x1a*\n\nBatchError\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05\x65rror\x18\x02 \x01(\tBo\n#io.weaviate.client.grpc.protocol.v1B\x12WeaviateProtoBatchZ4github.com/weaviate/weaviate/grpc/generated;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'v1.batch_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n#io.weaviate.client.grpc.protocol.v1B\022WeaviateProtoBatchZ4github.com/weaviate/weaviate/grpc/generated;protocol'
  _globals['_BATCHOBJECT'].fields_by_name['vector']._loaded_options = None
  _globals['_BATCHOBJECT'].fields_by_name['vector']._serialized_options = b'\030\001'
  _globals['_BATCHOBJECTSREQUEST']._serialized_start=77
  _globals['_BATCHOBJECTSREQUEST']._serialized_end=226
  _globals['_BATCHREFERENCESREQUEST']._serialized_start=229
  _globals['_BATCHREFERENCESREQUEST']._serialized_end=387
  _globals['_BATCHSENDREQUEST']._serialized_start=390
  _globals['_BATCHSENDREQUEST']._serialized_end=732
  _globals['_BATCHSENDREQUEST_STOP']._serialized_start=603
  _globals['_BATCHSENDREQUEST_STOP']._serialized_end=609
  _globals['_BATCHSENDREQUEST_OBJECTS']._serialized_start=611
  _globals['_BATCHSENDREQUEST_OBJECTS']._serialized_end=662
  _globals['_BATCHSENDREQUEST_REFERENCES']._serialized_start=664
  _globals['_BATCHSENDREQUEST_REFERENCES']._serialized_end=721
  _globals['_BATCHSENDREPLY']._serialized_start=734
  _globals['_BATCHSENDREPLY']._serialized_end=800
  _globals['_BATCHSTREAMREQUEST']._serialized_start=803
  _globals['_BATCHSTREAMREQUEST']._serialized_end=1002
  _globals['_BATCHSTREAMMESSAGE']._serialized_start=1005
  _globals['_BATCHSTREAMMESSAGE']._serialized_end=1501
  _globals['_BATCHSTREAMMESSAGE_START']._serialized_start=1345
  _globals['_BATCHSTREAMMESSAGE_START']._serialized_end=1352
  _globals['_BATCHSTREAMMESSAGE_STOP']._serialized_start=603
  _globals['_BATCHSTREAMMESSAGE_STOP']._serialized_end=609
  _globals['_BATCHSTREAMMESSAGE_SHUTDOWN']._serialized_start=1362
  _globals['_BATCHSTREAMMESSAGE_SHUTDOWN']._serialized_end=1372
  _globals['_BATCHSTREAMMESSAGE_SHUTTINGDOWN']._serialized_start=1374
  _globals['_BATCHSTREAMMESSAGE_SHUTTINGDOWN']._serialized_end=1388
  _globals['_BATCHSTREAMMESSAGE_ERROR']._serialized_start=1390
  _globals['_BATCHSTREAMMESSAGE_ERROR']._serialized_end=1490
  _globals['_BATCHOBJECT']._serialized_start=1504
  _globals['_BATCHOBJECT']._serialized_end=2494
  _globals['_BATCHOBJECT_PROPERTIES']._serialized_start=1708
  _globals['_BATCHOBJECT_PROPERTIES']._serialized_end=2352
  _globals['_BATCHOBJECT_SINGLETARGETREFPROPS']._serialized_start=2354
  _globals['_BATCHOBJECT_SINGLETARGETREFPROPS']._serialized_end=2410
  _globals['_BATCHOBJECT_MULTITARGETREFPROPS']._serialized_start=2412
  _globals['_BATCHOBJECT_MULTITARGETREFPROPS']._serialized_end=2494
  _globals['_BATCHREFERENCE']._serialized_start=2497
  _globals['_BATCHREFERENCE']._serialized_end=2650
  _globals['_BATCHOBJECTSREPLY']._serialized_start=2653
  _globals['_BATCHOBJECTSREPLY']._serialized_end=2789
  _globals['_BATCHOBJECTSREPLY_BATCHERROR']._serialized_start=2747
  _globals['_BATCHOBJECTSREPLY_BATCHERROR']._serialized_end=2789
  _globals['_BATCHREFERENCESREPLY']._serialized_start=2792
  _globals['_BATCHREFERENCESREPLY']._serialized_end=2934
  _globals['_BATCHREFERENCESREPLY_BATCHERROR']._serialized_start=2747
  _globals['_BATCHREFERENCESREPLY_BATCHERROR']._serialized_end=2789
# @@protoc_insertion_point(module_scope)



import json
from dataclasses import asdict
from typing import Generic, List, Literal, Optional, Type, Union, overload

from weaviate.cluster import _ClusterAsync
from weaviate.collections.aggregate import _AggregateCollectionAsync
from weaviate.collections.backups import _CollectionBackupAsync
from weaviate.collections.classes.cluster import Shard
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.grpc import METADATA, PROPERTIES, REFERENCES
from weaviate.collections.classes.internal import (
    CrossReferences,
    References,
    ReturnProperties,
    ReturnReferences,
    TReferences,
)
from weaviate.collections.classes.tenants import Tenant
from weaviate.collections.classes.types import Properties, TProperties
from weaviate.collections.config import _ConfigCollectionAsync
from weaviate.collections.data import _DataCollectionAsync
from weaviate.collections.generate import _GenerateCollectionAsync
from weaviate.collections.iterator import _IteratorInputs, _ObjectAIterator
from weaviate.collections.query import _QueryCollectionAsync
from weaviate.collections.tenants import _TenantsAsync
from weaviate.connect.v4 import ConnectionAsync
from weaviate.types import UUID

from .base import _CollectionBase


class CollectionAsync(Generic[Properties, References], _CollectionBase[ConnectionAsync]):
    """The collection class is the main entry point for interacting with a collection in Weaviate.

    This class is returned by the `client.collections.create` and `client.collections.get` methods. It provides
    access to all the methods available to you when interacting with a collection in Weaviate.

    You should not need to instantiate this class yourself but it may be useful to import this as a type when
    performing type hinting of functions that depend on a collection object.

    Attributes:
        aggregate: This namespace includes all the querying methods available to you when using Weaviate's standard aggregation capabilities.
        aggregate_group_by: This namespace includes all the aggregate methods available to you when using Weaviate's aggregation group-by capabilities.
        config: This namespace includes all the CRUD methods available to you when modifying the configuration of the collection in Weaviate.
        data: This namespace includes all the CUD methods available to you when modifying the data of the collection in Weaviate.
        generate: This namespace includes all the querying methods available to you when using Weaviate's generative capabilities.
        query_group_by: This namespace includes all the querying methods available to you when using Weaviate's querying group-by capabilities.
        query: This namespace includes all the querying methods available to you when using Weaviate's standard query capabilities.
        tenants: This namespace includes all the CRUD methods available to you when modifying the tenants of a multi-tenancy-enabled collection in Weaviate.
    """

    def __init__(
        self,
        connection: ConnectionAsync,
        name: str,
        validate_arguments: bool,
        consistency_level: Optional[ConsistencyLevel] = None,
        tenant: Optional[str] = None,
        properties: Optional[Type[Properties]] = None,
        references: Optional[Type[References]] = None,
    ) -> None:
        super().__init__(
            connection,
            name,
            validate_arguments,
            consistency_level,
            tenant,
        )
        self.__properties = properties
        self.__references = references

        self.__cluster = _ClusterAsync(connection)

        self.aggregate: _AggregateCollectionAsync = _AggregateCollectionAsync(
            connection, name, consistency_level, tenant, validate_arguments
        )
        """This namespace includes all the querying methods available to you when using Weaviate's standard aggregation capabilities."""
        self.backup: _CollectionBackupAsync = _CollectionBackupAsync(connection, name)
        """This namespace includes all the backup methods available to you when backing up a collection in Weaviate."""
        self.config = _ConfigCollectionAsync(connection, name, tenant)
        """This namespace includes all the CRUD methods available to you when modifying the configuration of the collection in Weaviate."""
        self.data = _DataCollectionAsync[Properties](
            connection, name, consistency_level, tenant, validate_arguments, properties
        )
        """This namespace includes all the CUD methods available to you when modifying the data of the collection in Weaviate."""
        self.generate: _GenerateCollectionAsync[Properties, References] = _GenerateCollectionAsync[
            Properties, References
        ](
            connection,
            name,
            consistency_level,
            tenant,
            properties,
            references,
            validate_arguments,
        )
        """This namespace includes all the querying methods available to you when using Weaviate's generative capabilities."""
        self.query = _QueryCollectionAsync[Properties, References](
            connection,
            name,
            consistency_level,
            tenant,
            properties,
            references,
            validate_arguments,
        )
        """This namespace includes all the querying methods available to you when using Weaviate's standard query capabilities."""
        self.tenants = _TenantsAsync(connection, name, validate_arguments)
        """This namespace includes all the CRUD methods available to you when modifying the tenants of a multi-tenancy-enabled collection in Weaviate."""

    def with_tenant(
        self, tenant: Union[str, Tenant, None]
    ) -> "CollectionAsync[Properties, References]":
        """Use this method to return a collection object specific to a single tenant.

        If multi-tenancy is not configured for this collection then Weaviate will throw an error.

        This method does not send a request to Weaviate. It only returns a new collection object that is specific
        to the tenant you specify.

        Args:
            tenant: The tenant to use. Can be `str` or `wvc.tenants.Tenant`.
        """
        return CollectionAsync(
            connection=self._connection,
            name=self.name,
            validate_arguments=self._validate_arguments,
            consistency_level=self.consistency_level,
            tenant=tenant.name if isinstance(tenant, Tenant) else tenant,
            properties=self.__properties,
            references=self.__references,
        )

    def with_consistency_level(
        self, consistency_level: Union[ConsistencyLevel, None]
    ) -> "CollectionAsync[Properties, References]":
        """Use this method to return a collection object specific to a single consistency level.

        If replication is not configured for this collection then Weaviate will throw an error.

        This method does not send a request to Weaviate. It only returns a new collection object that is specific
        to the consistency level you specify.

        Args:
            consistency_level: The consistency level to use.
        """
        return CollectionAsync(
            connection=self._connection,
            name=self.name,
            validate_arguments=self._validate_arguments,
            consistency_level=consistency_level,
            tenant=self.tenant,
            properties=self.__properties,
            references=self.__references,
        )

    async def length(self) -> int:
        """Get the total number of objects in the collection."""
        total = (await self.aggregate.over_all(total_count=True)).total_count
        assert total is not None
        return total

    async def to_string(self) -> str:
        """Return a string representation of the collection object."""
        config = await self.config.get()
        json_ = json.dumps(asdict(config), indent=2)
        return f"<weaviate.Collection config={json_}>"

    async def exists(self) -> bool:
        """Check if the collection exists in Weaviate."""
        try:
            await self.config.get(simple=True)
            return True
        except Exception:
            return False

    async def shards(self) -> List[Shard]:
        """Get the statuses of all the shards of this collection.

        Returns:
            The list of shards belonging to this collection.

        Raises:
            weaviate.exceptions.WeaviateConnectionError: If the network connection to weaviate fails.
            weaviate.exceptions.UnexpectedStatusCodeError: If weaviate reports a none OK status.
            weaviate.EmptyResponseError: If the response is empty.
        """
        return [
            shard
            for node in await self.__cluster.nodes(self.name, output="verbose")
            for shard in node.shards
        ]

    @overload
    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Optional[PROPERTIES] = None,
        return_references: Literal[None] = None,
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> _ObjectAIterator[Properties, References]: ...

    @overload
    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Optional[PROPERTIES] = None,
        return_references: REFERENCES,
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> _ObjectAIterator[Properties, CrossReferences]: ...

    @overload
    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Optional[PROPERTIES] = None,
        return_references: Type[TReferences],
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> _ObjectAIterator[Properties, TReferences]: ...

    @overload
    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Type[TProperties],
        return_references: Literal[None] = None,
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> _ObjectAIterator[TProperties, References]: ...

    @overload
    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Type[TProperties],
        return_references: REFERENCES,
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> _ObjectAIterator[TProperties, CrossReferences]: ...

    @overload
    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Type[TProperties],
        return_references: Type[TReferences],
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> _ObjectAIterator[TProperties, TReferences]: ...

    def iterator(
        self,
        include_vector: bool = False,
        return_metadata: Optional[METADATA] = None,
        *,
        return_properties: Optional[ReturnProperties[TProperties]] = None,
        return_references: Optional[ReturnReferences[TReferences]] = None,
        after: Optional[UUID] = None,
        cache_size: Optional[int] = None,
    ) -> Union[
        _ObjectAIterator[Properties, References],
        _ObjectAIterator[Properties, CrossReferences],
        _ObjectAIterator[Properties, TReferences],
        _ObjectAIterator[TProperties, References],
        _ObjectAIterator[TProperties, CrossReferences],
        _ObjectAIterator[TProperties, TReferences],
    ]:
        """Use this method to return an iterator over the objects in the collection.

        This iterator keeps a record of the last object that it returned to be used in each subsequent call to
        Weaviate. Once the collection is exhausted, the iterator exits.

        If `return_properties` is not provided, all the properties of each object will be
        requested from Weaviate except for its vector as this is an expensive operation. Specify `include_vector`
        to request the vector back as well. In addition, if `return_references=None` then none of the references
        are returned. Use `wvc.QueryReference` to specify which references to return.

        Args:
            include_vector: Whether to include the vector in the metadata of the returned objects.
            return_metadata: The metadata to return with each object.
            return_properties: The properties to return with each object.
            return_references: The references to return with each object.
            after: The cursor to use to mark the initial starting point of the iterator in the collection.
            cache_size: How many objects should be fetched in each request to Weaviate during the iteration. The default is 100.

        Raises:
            weaviate.exceptions.WeaviateGRPCQueryError: If the request to the Weaviate server fails.
        """
        return _ObjectAIterator(
            self.query,
            _IteratorInputs(
                include_vector=include_vector,
                return_metadata=return_metadata,
                return_properties=return_properties,
                return_references=return_references,
                after=after,
            ),
            cache_size=cache_size,
        )



from weaviate.collections.aggregations.near_vector.executor import _NearVectorExecutor
from weaviate.connect import executor
from weaviate.connect.v4 import ConnectionSync


@executor.wrap("sync")
class _NearVector(_NearVectorExecutor[ConnectionSync]):
    pass
